"""Enrichment domain entities."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime

# TODO(Phil): db stuff should not exist in the domain layer, create a mapper for this.
from kodit.infrastructure.sqlalchemy import entities as db_entities


@dataclass(frozen=True)
class EnrichmentV2(ABC):
    """Generic enrichment that can be attached to any entity."""

    content: str = ""
    id: int | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None

    @property
    @abstractmethod
    def type(self) -> str:
        """Return the enrichment type."""

    @property
    @abstractmethod
    def subtype(self) -> str | None:
        """Return the enrichment subtype (optional for hierarchical types)."""

    @abstractmethod
    def entity_type_key(self) -> str:
        """Return the entity type key this enrichment is for."""


@dataclass(frozen=True)
class CommitEnrichment(EnrichmentV2, ABC):
    """Enrichment specific to commits."""

    def entity_type_key(self) -> str:
        """Return the entity type key this enrichment is for."""
        return db_entities.GitCommit.__tablename__


@dataclass(frozen=True)
class EnrichmentAssociation:
    """Association between an enrichment and an entity."""

    enrichment_id: int
    entity_id: str
    entity_type: str
    id: int | None = None


@dataclass(frozen=True)
class CommitEnrichmentAssociation(EnrichmentAssociation):
    """Association between a commit and an enrichment."""

    entity_type: str = db_entities.GitCommit.__tablename__


@dataclass
class SnippetSummaryAssociation:
    """Association between a snippet and a summary enrichment."""

    snippet_summary: EnrichmentAssociation
    snippet: EnrichmentAssociation



"""Tests for the local embedding provider."""

from typing import Any
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from kodit.domain.value_objects import EmbeddingRequest
from kodit.infrastructure.embedding.embedding_providers.local_embedding_provider import (  # noqa: E501
    TINY,
    LocalEmbeddingProvider,
)


class TestLocalEmbeddingProvider:
    """Test the local embedding provider."""

    def test_init_default_values(self) -> None:
        """Test initialization with default values."""
        provider = LocalEmbeddingProvider()
        assert (
            provider.model_name
            == "flax-sentence-embeddings/st-codesearch-distilroberta-base"
        )
        assert provider.log is not None

    def test_init_custom_values(self) -> None:
        """Test initialization with custom values."""
        provider = LocalEmbeddingProvider(model_name=TINY)
        assert provider.model_name == "ibm-granite/granite-embedding-30m-english"

    def test_init_full_model_name(self) -> None:
        """Test initialization with a full model name."""
        custom_model = "sentence-transformers/all-MiniLM-L6-v2"
        provider = LocalEmbeddingProvider(model_name=custom_model)
        assert provider.model_name == custom_model

    @pytest.mark.asyncio
    async def test_embed_empty_requests(self) -> None:
        """Test embedding with empty requests."""
        provider = LocalEmbeddingProvider()
        requests: list[EmbeddingRequest] = []

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        assert len(results) == 0

    @pytest.mark.asyncio
    async def test_embed_single_request(self) -> None:
        """Test embedding with a single request."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model
            mock_model = MagicMock()
            mock_model.encode.return_value = np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [EmbeddingRequest(snippet_id="1", text="python programming")]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            assert len(results) == 1
            assert results[0].snippet_id == "1"
            assert len(results[0].embedding) == 5  # Based on our mock
            assert all(isinstance(v, float) for v in results[0].embedding)

    @pytest.mark.asyncio
    async def test_embed_multiple_requests(self) -> None:
        """Test embedding with multiple requests."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model
            mock_model = MagicMock()
            mock_model.encode.return_value = np.array(
                [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]
            )
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [
                EmbeddingRequest(snippet_id="1", text="python programming"),
                EmbeddingRequest(snippet_id="2", text="javascript development"),
                EmbeddingRequest(snippet_id="3", text="java enterprise"),
            ]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            assert len(results) == 3
            for i, result in enumerate(results):
                assert result.snippet_id == str(i + 1)
                assert len(result.embedding) == 3  # Based on our mock
                assert all(isinstance(v, float) for v in result.embedding)

    @pytest.mark.asyncio
    async def test_embed_different_texts_different_embeddings(self) -> None:
        """Test that different texts produce different embeddings."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model to return different embeddings for different texts
            mock_model = MagicMock()
            mock_model.encode.return_value = np.array(
                [
                    [0.1, 0.2, 0.3],
                    [0.9, 0.8, 0.7],  # Different embedding
                ]
            )
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [
                EmbeddingRequest(snippet_id="1", text="python programming"),
                EmbeddingRequest(snippet_id="2", text="javascript development"),
            ]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            assert len(results) == 2
            assert results[0].embedding != results[1].embedding

    @pytest.mark.asyncio
    async def test_embed_batch_processing(self) -> None:
        """Test that requests are processed in batches based on token limits."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model with dynamic response based on input size
            mock_model = MagicMock()

            def mock_encode(texts: list[str], **kwargs: Any) -> np.ndarray:  # noqa: ARG001
                # Return embeddings matching input size
                return np.array([[0.1, 0.2, 0.3] for _ in range(len(texts))])

            mock_model.encode = mock_encode
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            # Create requests with different token counts
            requests = [
                EmbeddingRequest(snippet_id=str(i), text=f"text {i}") for i in range(5)
            ]

            batch_count = 0
            total_results = []
            async for batch in provider.embed(requests):
                batch_count += 1
                total_results.extend(batch)

            assert len(total_results) == 5
            # Should be at least 1 batch, but exact count depends on tokenization

    @pytest.mark.asyncio
    async def test_embed_error_handling(self) -> None:
        """Test error handling during embedding."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model to raise an exception
            mock_model = MagicMock()
            mock_model.encode.side_effect = Exception("Model error")
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [EmbeddingRequest(snippet_id="1", text="test text")]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            # Should return no embeddings on error (empty batch)
            assert len(results) == 0

    @pytest.mark.asyncio
    async def test_embed_empty_text(self) -> None:
        """Test embedding with empty text."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model
            mock_model = MagicMock()
            mock_model.encode.return_value = np.array([[0.1, 0.2, 0.3]])
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [EmbeddingRequest(snippet_id="1", text="")]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            assert len(results) == 1
            assert all(isinstance(v, float) for v in results[0].embedding)

    @pytest.mark.asyncio
    async def test_embed_unicode_text(self) -> None:
        """Test embedding with unicode text."""
        with patch(
            "sentence_transformers.SentenceTransformer"
        ) as mock_transformer_class:
            # Mock the model
            mock_model = MagicMock()
            mock_model.encode.return_value = np.array([[0.1, 0.2, 0.3]])
            mock_transformer_class.return_value = mock_model

            provider = LocalEmbeddingProvider()
            requests = [EmbeddingRequest(snippet_id="1", text="python 🐍 programming")]

            results = []
            async for batch in provider.embed(requests):
                results.extend(batch)

            assert len(results) == 1
            assert all(isinstance(v, float) for v in results[0].embedding)

    def test_split_sub_batches(self) -> None:
        """Test the sub-batch splitting logic."""
        with patch("tiktoken.encoding_for_model") as mock_encoding_for_model:
            # Mock the encoding
            mock_encoding = MagicMock()
            mock_encoding.encode.side_effect = lambda text, **kwargs: [1] * len(  # noqa: ARG005
                text
            )  # Simple tokenization
            mock_encoding_for_model.return_value = mock_encoding

            provider = LocalEmbeddingProvider()

            # Test with short texts that should fit in one batch
            short_requests = [
                EmbeddingRequest(snippet_id="1", text="short"),
                EmbeddingRequest(snippet_id="2", text="text"),
            ]

            batches = provider._split_sub_batches(mock_encoding, short_requests)  # noqa: SLF001
            assert len(batches) == 1
            assert len(batches[0]) == 2

            # Test with long text that should be split
            long_requests = [
                EmbeddingRequest(snippet_id="1", text="x" * 10000),  # Very long text
            ]

            batches = provider._split_sub_batches(mock_encoding, long_requests)  # noqa: SLF001
            assert len(batches) == 1  # Single item should still be in one batch
            assert len(batches[0]) == 1



"""Tests for CookbookContextService."""

import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest

from kodit.domain.services.cookbook_context_service import CookbookContextService


@pytest.fixture
def temp_repo() -> Generator[Path, None, None]:
    """Create a temporary repository directory with basic files."""
    with tempfile.TemporaryDirectory() as tmp:
        repo_path = Path(tmp)

        # Create README
        (repo_path / "README.md").write_text(
            "# Test Library\nA simple testing library for demos."
        )

        # Create package.json
        (repo_path / "package.json").write_text(
            '{"name": "test-lib", "version": "1.0.0"}'
        )

        yield repo_path


@pytest.fixture
def sample_api_modules() -> list[Any]:
    """Create sample API module definitions using mocks."""
    # Create mock function
    func = MagicMock()
    func.simple_name = "create_client"
    func.docstring = "Create a new client instance"

    # Create mock method
    method = MagicMock()
    method.simple_name = "connect"
    method.docstring = "Connect to server"

    # Create mock class
    cls = MagicMock()
    cls.simple_name = "Client"
    cls.docstring = "Main client class"
    cls.methods = [method]

    # Create mock module
    module = MagicMock()
    module.module_path = "test_lib.core"
    module.module_docstring = "Core functionality"
    module.functions = [func]
    module.classes = [cls]
    module.types = []

    return [module]


@pytest.mark.asyncio
async def test_gather_context_includes_all_sections(
    temp_repo: Path, sample_api_modules: list[Any]
) -> None:
    """Test that gather_context includes all major sections."""
    service = CookbookContextService()

    context = await service.gather_context(
        temp_repo, language="javascript", api_modules=sample_api_modules
    )

    # Verify all expected sections are present
    assert "## Primary Language" in context
    assert "javascript" in context
    assert "## README" in context
    assert "Test Library" in context
    assert "## API Structure" in context
    assert "test_lib.core" in context
    assert "## Package Information" in context
    assert "package.json" in context


@pytest.mark.asyncio
async def test_gather_context_without_api_modules(temp_repo: Path) -> None:
    """Test that gather_context works without API modules."""
    service = CookbookContextService()

    context = await service.gather_context(temp_repo, language="python")

    # Should still have basic sections
    assert "## Primary Language" in context
    assert "python" in context
    assert "## README" in context
    # Should not have API structure section
    assert "## API Structure" not in context


@pytest.mark.asyncio
async def test_format_api_structure_includes_classes_and_functions(
    sample_api_modules: list[Any],
) -> None:
    """Test that API structure formatting includes key elements."""
    service = CookbookContextService()

    formatted = service._format_api_structure(sample_api_modules)  # noqa: SLF001

    # Verify module is listed
    assert "test_lib.core" in formatted
    # Verify class is listed
    assert "Client" in formatted
    # Verify function is listed
    assert "create_client" in formatted
    # Verify method is listed
    assert "connect" in formatted


@pytest.mark.asyncio
async def test_format_api_structure_with_empty_modules() -> None:
    """Test that empty modules list returns empty string."""
    service = CookbookContextService()

    formatted = service._format_api_structure([])  # noqa: SLF001

    assert formatted == ""



"""Queue service for managing tasks."""

from collections.abc import Callable
from typing import Any

import structlog
from sqlalchemy.ext.asyncio import AsyncSession

from kodit.domain.entities import Task
from kodit.domain.value_objects import QueuePriority, TaskOperation
from kodit.infrastructure.sqlalchemy.query import FilterOperator, QueryBuilder
from kodit.infrastructure.sqlalchemy.task_repository import (
    create_task_repository,
)


class QueueService:
    """Service for queue operations using database persistence.

    This service provides the main interface for enqueuing and managing tasks.
    It uses the existing Task entity in the database with a flexible JSON payload.
    """

    def __init__(
        self,
        session_factory: Callable[[], AsyncSession],
    ) -> None:
        """Initialize the queue service."""
        self.task_repository = create_task_repository(session_factory=session_factory)
        self.log = structlog.get_logger(__name__)

    async def enqueue_task(self, task: Task) -> None:
        """Queue a task in the database."""
        # Check if task already exists
        exists = await self.task_repository.exists(task.id)
        await self.task_repository.save(task)
        if exists:
            self.log.info("Task updated", task_id=task.id, task_type=task.type)
        else:
            self.log.info(
                "Task queued",
                task_id=task.id,
                task_type=task.type,
                payload=task.payload,
            )

    async def enqueue_tasks(
        self,
        tasks: list[TaskOperation],
        base_priority: QueuePriority,
        payload: dict[str, Any],
    ) -> None:
        """Queue repository tasks."""
        priority_offset = len(tasks) * 10
        for task in tasks:
            await self.enqueue_task(
                Task.create(
                    task,
                    base_priority + priority_offset,
                    payload,
                )
            )
            priority_offset -= 10

    async def list_tasks(
        self, task_operation: TaskOperation | None = None
    ) -> list[Task]:
        """List all tasks in the queue."""
        query = QueryBuilder()
        if task_operation:
            query.filter("type", FilterOperator.EQ, task_operation.value)
        query.sort("priority", descending=True)
        query.sort("created_at", descending=True)
        return await self.task_repository.find(query)

    async def get_task(self, task_id: str) -> Task | None:
        """Get a specific task by ID."""
        try:
            return await self.task_repository.get(task_id)
        except ValueError:
            return None



"""Complete self-contained analyzer for kodit-slicer.

This module combines all necessary functionality without external dependencies
on the legacy domain/application/infrastructure layers.
"""

from collections import defaultdict
from collections.abc import Generator
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any

import structlog
from tree_sitter import Node, Parser, Tree

from kodit.domain.entities.git import GitFile, SnippetV2
from kodit.domain.value_objects import LanguageMapping
from kodit.infrastructure.slicing.ast_analyzer import ASTAnalyzer
from kodit.infrastructure.slicing.code_elements import FunctionInfo
from kodit.infrastructure.slicing.language_analyzer import language_analyzer_factory

if TYPE_CHECKING:
    from kodit.infrastructure.slicing.code_elements import (
        FunctionDefinition,
        ParsedFile,
        TypeDefinition,
    )


@dataclass
class AnalyzerState:
    """Central state for the dependency analysis."""

    parser: Parser
    files: list[Path] = field(default_factory=list)
    asts: dict[Path, Tree] = field(default_factory=dict)
    def_index: dict[str, FunctionInfo] = field(default_factory=dict)
    call_graph: dict[str, set[str]] = field(default_factory=lambda: defaultdict(set))
    reverse_calls: dict[str, set[str]] = field(default_factory=lambda: defaultdict(set))
    imports: dict[Path, dict[str, str]] = field(
        default_factory=lambda: defaultdict(dict)
    )


class Slicer:
    """Slicer that extracts code snippets from files."""

    def __init__(self) -> None:
        """Initialize an empty slicer."""
        self.log = structlog.get_logger(__name__)

    def extract_snippets_from_git_files(  # noqa: C901
        self, files: list[GitFile], language: str = "python"
    ) -> list[SnippetV2]:
        """Extract code snippets from a list of files.

        Args:
            files: List of domain File objects to analyze
            language: Programming language for analysis

        Returns:
            List of extracted code snippets as domain entities

        Raises:
            FileNotFoundError: If any file doesn't exist

        """
        if not files:
            return []

        language = language.lower()

        # Initialize ASTAnalyzer
        try:
            analyzer = ASTAnalyzer(language)
        except ValueError:
            self.log.debug("Skipping unsupported language", language=language)
            return []

        # Validate files
        path_to_file_map: dict[Path, GitFile] = {}
        for file in files:
            file_path = Path(file.path)

            # Validate file matches language
            if not self._file_matches_language(file_path.suffix, language):
                raise ValueError(f"File {file_path} does not match language {language}")

            # Validate file exists
            if not file_path.exists():
                raise FileNotFoundError(f"File not found: {file_path}")

            path_to_file_map[file_path] = file

        # Parse files and extract definitions using ASTAnalyzer
        parsed_files = analyzer.parse_files(files)
        if not parsed_files:
            return []

        functions, _, types = analyzer.extract_definitions(
            parsed_files, include_private=True
        )

        # Build state from ASTAnalyzer results
        state = self._build_state_from_ast_analyzer(parsed_files, functions)
        lang_analyzer = language_analyzer_factory(language)

        # Build type lookup by simple name for prepending to function snippets
        type_lookup: dict[str, TypeDefinition] = {t.simple_name: t for t in types}

        # Build call graph and snippets (Slicer-specific logic)
        self._build_call_graph(state, lang_analyzer)
        self._build_reverse_call_graph(state)

        # Extract snippets for all functions
        file_contents: dict[Path, str] = {}
        snippets: list[SnippetV2] = []
        for qualified_name in state.def_index:
            snippet_content = self._get_snippet(
                qualified_name,
                state,
                file_contents,
                {"max_depth": 2, "max_functions": 8},
            )
            if "not found" not in snippet_content:
                # Prepend referenced types for TypeScript/TSX
                if language in ("typescript", "ts", "tsx"):
                    snippet_content = self._prepend_referenced_types(
                        qualified_name,
                        snippet_content,
                        state,
                        type_lookup,
                        file_contents,
                        lang_analyzer,
                    )

                snippet = self._create_snippet_entity_from_git_files(
                    qualified_name, snippet_content, language, state, path_to_file_map
                )
                snippets.append(snippet)

        # Extract snippets for types (interfaces, type aliases)
        for type_def in types:
            type_snippet = self._create_type_snippet(
                type_def, language, path_to_file_map, file_contents
            )
            if type_snippet:
                snippets.append(type_snippet)

        # Extract JSX snippets from functions with large JSX returns
        jsx_snippets = self._extract_jsx_snippets(
            state, language, path_to_file_map, file_contents
        )
        snippets.extend(jsx_snippets)

        # Extract top-level entry point expressions (e.g., ReactDOM.render calls)
        entry_point_snippets = self._extract_entry_point_snippets(
            parsed_files, language, path_to_file_map, file_contents
        )
        snippets.extend(entry_point_snippets)

        return snippets

    def _file_matches_language(self, file_extension: str, language: str) -> bool:
        """Check if a file extension matches the current language."""
        try:
            language_analyzer_factory(language)
        except ValueError:
            return False

        try:
            return language == LanguageMapping.get_language_for_extension(
                file_extension
            )
        except ValueError:
            # Extension not supported, so it doesn't match any language
            return False

    def _build_state_from_ast_analyzer(
        self,
        parsed_files: list["ParsedFile"],
        functions: list["FunctionDefinition"],
    ) -> AnalyzerState:
        """Build AnalyzerState from ASTAnalyzer results."""
        # Create a dummy parser (not used for new parsing)
        from tree_sitter_language_pack import get_language

        ts_language = get_language("python")
        parser = Parser(ts_language)

        state = AnalyzerState(parser=parser)

        # Populate files and ASTs from ParsedFile objects
        for parsed in parsed_files:
            state.files.append(parsed.path)
            state.asts[parsed.path] = parsed.tree

        # Populate def_index from FunctionDefinition objects
        for func_def in functions:
            state.def_index[func_def.qualified_name] = FunctionInfo(
                file=func_def.file,
                node=func_def.node,
                span=func_def.span,
                qualified_name=func_def.qualified_name,
            )

        return state

    def _build_call_graph(self, state: AnalyzerState, analyzer: Any) -> None:
        """Build call graph from function definitions."""
        for qualified_name, func_info in state.def_index.items():
            calls = self._find_function_calls(
                func_info.node, func_info.file, state, analyzer
            )
            state.call_graph[qualified_name] = calls

    def _build_reverse_call_graph(self, state: AnalyzerState) -> None:
        """Build reverse call graph."""
        for caller, callees in state.call_graph.items():
            for callee in callees:
                state.reverse_calls[callee].add(caller)

    def _walk_tree(self, node: Node) -> Generator[Node, None, None]:
        """Walk the AST tree, yielding all nodes."""
        # Use a simple queue-based approach to avoid recursion issues
        queue = [node]
        visited: set[int] = set()  # Track by node id (memory address)

        while queue:
            current = queue.pop(0)

            # Use node id (memory address) as unique identifier to avoid infinite loops
            node_id = id(current)
            if node_id in visited:
                continue
            visited.add(node_id)

            yield current

            # Add children to queue
            queue.extend(current.children)

    def _get_file_content(self, file_path: Path, file_contents: dict[Path, str]) -> str:
        """Get cached file content."""
        if file_path not in file_contents:
            try:
                with file_path.open(encoding="utf-8") as f:
                    file_contents[file_path] = f.read()
            except UnicodeDecodeError as e:
                file_contents[file_path] = f"# Error reading file: {e}"
            except OSError as e:
                file_contents[file_path] = f"# Error reading file: {e}"
        return file_contents[file_path]

    def _get_snippet(
        self,
        function_name: str,
        state: AnalyzerState,
        file_contents: dict[Path, str],
        snippet_config: dict[str, Any] | None = None,
    ) -> str:
        """Generate a smart snippet for a function with its dependencies."""
        if snippet_config is None:
            snippet_config = {}

        max_depth = snippet_config.get("max_depth", 2)
        max_functions = snippet_config.get("max_functions", 8)
        include_usage = snippet_config.get("include_usage", True)

        if function_name not in state.def_index:
            return f"Error: Function '{function_name}' not found"

        # Find dependencies
        dependencies = self._find_dependencies(
            function_name, state, max_depth, max_functions
        )

        # Filter out nested functions (they're already separate snippets)
        func_info = state.def_index[function_name]
        nested_func_names = {
            name
            for name, info in state.def_index.items()
            if (
                info.file == func_info.file
                and info.span[0] > func_info.span[0]
                and info.span[1] <= func_info.span[1]
            )
        }
        dependencies = dependencies - nested_func_names

        # Sort dependencies topologically
        sorted_deps = self._topological_sort(dependencies, state)

        # Build snippet
        snippet_lines = []

        # Add imports
        imports = self._get_minimal_imports({function_name}.union(dependencies))
        if imports:
            snippet_lines.extend(imports)
            snippet_lines.append("")

        # Add target function
        target_source = self._extract_function_source(
            function_name, state, file_contents
        )
        snippet_lines.append(target_source)

        # Add dependencies (excluding nested functions)
        if dependencies:
            snippet_lines.append("")
            snippet_lines.append("# === DEPENDENCIES ===")
            for dep in sorted_deps:
                snippet_lines.append("")
                dep_source = self._extract_function_source(dep, state, file_contents)
                snippet_lines.append(dep_source)

        # Add usage examples
        if include_usage:
            callers = state.reverse_calls.get(function_name, set())
            if callers:
                snippet_lines.append("")
                snippet_lines.append("# === USAGE EXAMPLES ===")
                # Show up to 2 examples, sorted for deterministic order
                for caller in sorted(callers)[:2]:
                    call_line = self._find_function_call_line(
                        caller, function_name, state, file_contents
                    )
                    if call_line and not call_line.startswith("#"):
                        snippet_lines.append(f"# From {caller}:")
                        snippet_lines.append(f"    {call_line}")
                        snippet_lines.append("")

        return "\n".join(snippet_lines)

    def _create_snippet_entity_from_git_files(
        self,
        qualified_name: str,
        snippet_content: str,
        language: str,
        state: AnalyzerState,
        path_to_file_map: dict[Path, GitFile],
    ) -> SnippetV2:
        """Create a Snippet domain entity from extracted content."""
        # Determine all files that this snippet derives from
        derives_from_files = self._find_source_files_for_snippet_from_git_files(
            qualified_name, snippet_content, state, path_to_file_map
        )

        # Create the snippet entity
        return SnippetV2(
            derives_from=derives_from_files,
            content=snippet_content,
            extension=language,
            sha=SnippetV2.compute_sha(snippet_content),
        )

    def _create_type_snippet(
        self,
        type_def: "TypeDefinition",
        language: str,
        path_to_file_map: dict[Path, GitFile],
        file_contents: dict[Path, str],
    ) -> SnippetV2 | None:
        """Create a snippet from a type definition (interface, type alias)."""
        # Get the source content for this type
        file_content = self._get_file_content(type_def.file, file_contents)
        source_bytes = file_content.encode("utf-8")
        start_byte, end_byte = type_def.span
        type_source = source_bytes[start_byte:end_byte].decode("utf-8")

        if not type_source.strip():
            return None

        # Find the source file
        derives_from: list[GitFile] = []
        if type_def.file in path_to_file_map:
            derives_from.append(path_to_file_map[type_def.file])

        return SnippetV2(
            derives_from=derives_from,
            content=type_source,
            extension=language,
            sha=SnippetV2.compute_sha(type_source),
        )

    def _extract_jsx_snippets(
        self,
        state: AnalyzerState,
        language: str,
        path_to_file_map: dict[Path, GitFile],
        file_contents: dict[Path, str],
    ) -> list[SnippetV2]:
        """Extract JSX return statements as separate snippets."""
        jsx_snippets: list[SnippetV2] = []
        min_lines = 5  # Only extract JSX returns with at least this many lines

        for qualified_name, func_info in state.def_index.items():
            file_content = self._get_file_content(func_info.file, file_contents)
            source_bytes = file_content.encode("utf-8")
            start_byte, end_byte = func_info.span
            func_source = source_bytes[start_byte:end_byte].decode("utf-8")

            # Look for JSX return statements
            jsx_content = self._extract_jsx_return_content(func_source, min_lines)
            if jsx_content:
                # Create snippet for the JSX
                simple_name = qualified_name.split(".")[-1]
                jsx_header = f"// JSX from {simple_name}\n"

                derives_from: list[GitFile] = []
                if func_info.file in path_to_file_map:
                    derives_from.append(path_to_file_map[func_info.file])

                jsx_snippets.append(
                    SnippetV2(
                        derives_from=derives_from,
                        content=jsx_header + jsx_content,
                        extension=language,
                        sha=SnippetV2.compute_sha(jsx_header + jsx_content),
                    )
                )

        return jsx_snippets

    def _extract_entry_point_snippets(
        self,
        parsed_files: list["ParsedFile"],
        language: str,
        path_to_file_map: dict[Path, GitFile],
        file_contents: dict[Path, str],
    ) -> list[SnippetV2]:
        """Extract top-level entry point expressions (e.g., ReactDOM.render calls)."""
        entry_snippets: list[SnippetV2] = []

        for parsed in parsed_files:
            file_content = self._get_file_content(parsed.path, file_contents)
            source_bytes = file_content.encode("utf-8")

            # Look for top-level expression statements
            for child in parsed.tree.root_node.children:
                if child.type == "expression_statement":
                    # Check if this is a meaningful entry point (render call, etc.)
                    expr_text = source_bytes[child.start_byte : child.end_byte].decode(
                        "utf-8"
                    )

                    # Skip trivial expressions (single identifiers, etc.)
                    if self._is_entry_point_expression(expr_text):
                        content = expr_text

                        derives_from: list[GitFile] = []
                        if parsed.path in path_to_file_map:
                            derives_from.append(path_to_file_map[parsed.path])

                        entry_snippets.append(
                            SnippetV2(
                                derives_from=derives_from,
                                content=content,
                                extension=language,
                                sha=SnippetV2.compute_sha(content),
                            )
                        )

        return entry_snippets

    def _is_entry_point_expression(self, expr_text: str) -> bool:
        """Check if an expression statement is a meaningful entry point."""
        import re

        # Check for common entry point patterns
        entry_patterns = [
            r"ReactDOM\.(createRoot|render)",  # React entry points
            r"render\s*\(",  # Generic render calls
            r"mount\s*\(",  # Vue/other framework mounts
            r"bootstrap\s*\(",  # Bootstrap patterns
            r"createApp\s*\(",  # Vue 3
        ]

        for pattern in entry_patterns:
            if re.search(pattern, expr_text):
                return True

        # Also check if expression contains JSX (indicates UI entry point)
        return bool(re.search(r"<\w+", expr_text))

    def _extract_jsx_return_content(
        self, func_source: str, min_lines: int = 5
    ) -> str | None:
        """Extract the JSX content from a return statement if it's large enough."""
        import re

        lines = func_source.split("\n")
        i = 0

        while i < len(lines):
            line = lines[i]
            stripped = line.strip()

            # Check if this is the start of a return statement
            if stripped.startswith("return (") or stripped == "return (":
                # Find the matching closing paren
                paren_count = line.count("(") - line.count(")")
                return_lines = [line]

                j = i + 1
                while j < len(lines) and paren_count > 0:
                    return_lines.append(lines[j])
                    paren_count += lines[j].count("(") - lines[j].count(")")
                    j += 1

                # Check if return contains JSX (has < tag)
                return_content = "\n".join(return_lines)
                has_jsx = bool(re.search(r"<\w+", return_content))

                # If it's a large JSX return, extract it
                if has_jsx and len(return_lines) >= min_lines:
                    return return_content

            i += 1

        return None

    def _find_source_files_for_snippet_from_git_files(
        self,
        qualified_name: str,
        snippet_content: str,
        state: AnalyzerState,
        path_to_file_map: dict[Path, GitFile],
    ) -> list[GitFile]:
        """Find all source files that a snippet derives from."""
        source_files: list[GitFile] = []
        source_file_paths: set[Path] = set()

        # Add the primary function's file
        if qualified_name in state.def_index:
            primary_file_path = state.def_index[qualified_name].file
            if (
                primary_file_path in path_to_file_map
                and primary_file_path not in source_file_paths
            ):
                source_files.append(path_to_file_map[primary_file_path])
                source_file_paths.add(primary_file_path)

        # Find all dependencies mentioned in the snippet and add their source files
        dependencies = self._extract_dependency_names_from_snippet(
            snippet_content, state
        )
        for dep_name in dependencies:
            if dep_name in state.def_index:
                dep_file_path = state.def_index[dep_name].file
                if (
                    dep_file_path in path_to_file_map
                    and dep_file_path not in source_file_paths
                ):
                    source_files.append(path_to_file_map[dep_file_path])
                    source_file_paths.add(dep_file_path)

        return source_files

    def _extract_dependency_names_from_snippet(
        self, snippet_content: str, state: AnalyzerState
    ) -> set[str]:
        """Extract dependency function names from snippet content."""
        dependencies: set[str] = set()

        # Look for the DEPENDENCIES section and extract function names
        lines = snippet_content.split("\n")
        in_dependencies_section = False

        for original_line in lines:
            line = original_line.strip()
            if line == "# === DEPENDENCIES ===":
                in_dependencies_section = True
                continue
            if line == "# === USAGE EXAMPLES ===":
                in_dependencies_section = False
                continue

            if in_dependencies_section and line.startswith("def "):
                # Extract function name from "def function_name(...)" pattern
                func_def_start = line.find("def ") + 4
                func_def_end = line.find("(", func_def_start)
                if func_def_end > func_def_start:
                    func_name = line[func_def_start:func_def_end].strip()
                    # Try to find the qualified name (module.function_name format)
                    # We need to search through the state.def_index to find matches
                    for qualified_name in self._get_qualified_names_for_function(
                        func_name, state
                    ):
                        dependencies.add(qualified_name)

        return dependencies

    def _get_qualified_names_for_function(
        self, func_name: str, state: AnalyzerState
    ) -> list[str]:
        """Get possible qualified names for a function name."""
        # This is a simple implementation - in practice you might want more
        # sophisticated matching
        return [
            qualified
            for qualified in state.def_index
            if qualified.endswith(f".{func_name}")
        ]

    # Helper methods

    def _extract_imports(self, node: Node) -> dict[str, str]:
        """Extract imports from import node."""
        imports = {}
        if node.type == "import_statement":
            for child in node.children:
                if child.type == "dotted_name" and child.text is not None:
                    module_name = child.text.decode("utf-8")
                    imports[module_name] = module_name
        elif node.type == "import_from_statement":
            module_node = node.child_by_field_name("module_name")
            if module_node and module_node.text is not None:
                module_name = module_node.text.decode("utf-8")
                for child in node.children:
                    if child.type == "import_list":
                        for import_child in child.children:
                            if (
                                import_child.type == "dotted_name"
                                and import_child.text is not None
                            ):
                                imported_name = import_child.text.decode("utf-8")
                                imports[imported_name] = (
                                    f"{module_name}.{imported_name}"
                                )
        return imports

    def _find_function_calls(
        self, node: Node, file_path: Path, state: AnalyzerState, analyzer: Any
    ) -> set[str]:
        """Find function calls in a node."""
        calls = set()
        call_node_type = analyzer.node_types().call_node

        for child in self._walk_tree(node):
            if child.type == call_node_type:
                function_node = child.child_by_field_name("function")
                if function_node:
                    call_name = self._extract_call_name(function_node)
                    if call_name:
                        resolved = self._resolve_call(call_name, file_path, state)
                        if resolved:
                            calls.add(resolved)
        return calls

    def _extract_call_name(self, node: Node) -> str | None:
        """Extract function name from call node."""
        if node.type == "identifier" and node.text is not None:
            return node.text.decode("utf-8")
        if node.type == "attribute":
            object_node = node.child_by_field_name("object")
            attribute_node = node.child_by_field_name("attribute")
            if (
                object_node
                and attribute_node
                and object_node.text is not None
                and attribute_node.text is not None
            ):
                obj_name = object_node.text.decode("utf-8")
                attr_name = attribute_node.text.decode("utf-8")
                return f"{obj_name}.{attr_name}"
        return None

    def _resolve_call(
        self, call_name: str, file_path: Path, state: AnalyzerState
    ) -> str | None:
        """Resolve a function call to qualified name."""
        module_name = file_path.stem
        local_qualified = f"{module_name}.{call_name}"

        if local_qualified in state.def_index:
            return local_qualified

        # Check imports
        if file_path in state.imports:
            imports = state.imports[file_path]
            if call_name in imports:
                return imports[call_name]

        # Check if already qualified
        if call_name in state.def_index:
            return call_name

        return None

    def _find_dependencies(
        self, target: str, state: AnalyzerState, max_depth: int, max_functions: int
    ) -> set[str]:
        """Find relevant dependencies for a function."""
        visited: set[str] = set()
        to_visit = [(target, 0)]
        dependencies: set[str] = set()

        while to_visit and len(dependencies) < max_functions:
            current, depth = to_visit.pop(0)
            if current in visited or depth > max_depth:
                continue
            visited.add(current)

            if current != target:
                dependencies.add(current)

            # Add direct dependencies
            to_visit.extend(
                (callee, depth + 1)
                for callee in sorted(state.call_graph.get(current, set()))
                if callee not in visited and callee in state.def_index
            )

        return dependencies

    def _topological_sort(self, functions: set[str], state: AnalyzerState) -> list[str]:
        """Sort functions in dependency order."""
        if not functions:
            return []

        # Build subgraph
        in_degree: dict[str, int] = defaultdict(int)
        graph: dict[str, set[str]] = defaultdict(set)

        for func in sorted(functions):
            for callee in sorted(state.call_graph.get(func, set())):
                if callee in functions:
                    graph[func].add(callee)
                    in_degree[callee] += 1

        # Find roots
        queue = [f for f in sorted(functions) if in_degree[f] == 0]
        result = []

        while queue:
            current = queue.pop(0)
            result.append(current)
            for neighbor in sorted(graph[current]):
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    queue.append(neighbor)

        # Add any remaining (cycles)
        for func in sorted(functions):
            if func not in result:
                result.append(func)

        return result

    def _get_minimal_imports(self, _functions: set[str]) -> list[str]:
        """Get minimal imports needed for functions."""
        # For now, we'll skip imports to simplify the refactoring
        return []

    def _extract_function_source(
        self, qualified_name: str, state: AnalyzerState, file_contents: dict[Path, str]
    ) -> str:
        """Extract complete function source code, summarizing nested functions."""
        if qualified_name not in state.def_index:
            return f"# Function {qualified_name} not found"

        func_info = state.def_index[qualified_name]
        file_content = self._get_file_content(func_info.file, file_contents)
        source_bytes = file_content.encode("utf-8")

        # Extract function source using byte positions
        start_byte, end_byte = func_info.span
        func_source = source_bytes[start_byte:end_byte].decode("utf-8")

        # Find nested functions that are contained within this function's span
        nested_functions = self._find_nested_functions(
            func_info, state, source_bytes, start_byte
        )

        # If there are nested functions, summarize them
        if nested_functions:
            func_source = self._summarize_nested_functions(
                func_source, nested_functions, start_byte
            )

        # Summarize large JSX return statements
        return self._summarize_jsx_return(func_source)


    def _find_nested_functions(
        self,
        parent_func: FunctionInfo,
        state: AnalyzerState,
        source_bytes: bytes,
        parent_start: int,
    ) -> list[tuple[str, int, int, str]]:
        """Find functions nested within a parent function.

        Returns list of (name, relative_start, relative_end, signature) tuples.
        """
        parent_start_byte, parent_end_byte = parent_func.span
        nested = []

        for name, info in state.def_index.items():
            # Skip self
            if name == parent_func.qualified_name:
                continue

            # Check if this function is nested within the parent
            if (
                info.file == parent_func.file
                and info.span[0] > parent_start_byte
                and info.span[1] <= parent_end_byte
            ):
                # Get the signature (first line up to the opening brace or arrow)
                func_bytes = source_bytes[info.span[0] : info.span[1]]
                func_text = func_bytes.decode("utf-8")
                signature = self._extract_function_signature(func_text)

                # Store relative positions within parent function
                rel_start = info.span[0] - parent_start
                rel_end = info.span[1] - parent_start
                nested.append((name.split(".")[-1], rel_start, rel_end, signature))

        # Sort by position (descending) so we can replace from end to start
        nested.sort(key=lambda x: x[1], reverse=True)
        return nested

    def _extract_function_signature(self, func_text: str) -> str:
        """Extract the signature part of a function (before the body)."""
        # For arrow functions: const foo = (args) => { ... }
        # We want: const foo = (args) =>
        if "=>" in func_text:
            arrow_pos = func_text.find("=>")
            return func_text[: arrow_pos + 2].strip()

        # For regular functions: function foo(args) { ... }
        # We want: function foo(args)
        if "{" in func_text:
            brace_pos = func_text.find("{")
            return func_text[:brace_pos].strip()

        return func_text.split("\n")[0].strip()

    def _summarize_nested_functions(
        self,
        func_source: str,
        nested_functions: list[tuple[str, int, int, str]],
        _parent_start: int,
    ) -> str:
        """Replace nested function bodies with '{ ... }' placeholders."""
        result = func_source

        for _name, rel_start, rel_end, signature in nested_functions:
            # Replace the entire nested function with just its signature + ...
            nested_text = result[rel_start:rel_end]

            # Preserve the indentation
            line_start = result.rfind("\n", 0, rel_start)
            if line_start != -1:
                line_content = result[line_start + 1 : rel_start]
                line_content[: len(line_content) - len(line_content.lstrip())]

            # Check if it ends with semicolon (for const declarations)
            ends_with_semi = nested_text.rstrip().endswith(";")
            suffix = ";" if ends_with_semi else ""

            # Create summarized version
            summarized = f"{signature} {{ ... }}{suffix}"
            result = result[:rel_start] + summarized + result[rel_end:]

        return result

    def _summarize_jsx_return(self, func_source: str, min_lines: int = 5) -> str:
        """Summarize large JSX return statements to 'return ( ... );'."""
        import re

        # Look for return statements with JSX (parenthesized expressions with <)
        # Pattern: return ( ... ); or return ( ... )
        # We need to handle multi-line returns with JSX

        lines = func_source.split("\n")
        result_lines = []
        i = 0

        while i < len(lines):
            line = lines[i]
            stripped = line.strip()

            # Check if this is the start of a return statement
            if stripped.startswith("return (") or stripped == "return (":
                # Find the matching closing paren
                paren_count = line.count("(") - line.count(")")
                return_lines = [line]

                j = i + 1
                while j < len(lines) and paren_count > 0:
                    return_lines.append(lines[j])
                    paren_count += lines[j].count("(") - lines[j].count(")")
                    j += 1

                # Check if return contains JSX (has < tag)
                return_content = "\n".join(return_lines)
                has_jsx = bool(re.search(r"<\w+", return_content))

                # If it's a large JSX return, summarize it
                if has_jsx and len(return_lines) >= min_lines:
                    # Get the indentation from the return line
                    indent = line[: len(line) - len(line.lstrip())]
                    result_lines.append(f"{indent}return ( ... );")
                    i = j
                else:
                    result_lines.append(line)
                    i += 1
            else:
                result_lines.append(line)
                i += 1

        return "\n".join(result_lines)

    def _find_function_call_line(
        self,
        caller_qualified_name: str,
        target_name: str,
        state: AnalyzerState,
        file_contents: dict[Path, str],
    ) -> str:
        """Find the actual line where a function calls another."""
        if caller_qualified_name not in state.def_index:
            return f"# calls {target_name}"

        caller_info = state.def_index[caller_qualified_name]
        file_content = self._get_file_content(caller_info.file, file_contents)
        source_bytes = file_content.encode("utf-8")

        # Extract the caller function source
        start_byte, end_byte = caller_info.span
        function_source = source_bytes[start_byte:end_byte].decode("utf-8")

        # Look for lines that contain the target function call
        lines = function_source.split("\n")
        target_simple_name = target_name.split(".")[-1]  # Get just the function name

        for line in lines:
            if target_simple_name in line and "(" in line:
                # Clean up the line (remove leading/trailing whitespace)
                clean_line = line.strip()
                if clean_line:
                    return clean_line

        return f"# calls {target_name}"

    def _prepend_referenced_types(  # noqa: PLR0913
        self,
        qualified_name: str,
        snippet_content: str,
        state: AnalyzerState,
        type_lookup: dict[str, "TypeDefinition"],
        file_contents: dict[Path, str],
        lang_analyzer: Any,
    ) -> str:
        """Prepend referenced type definitions to a function snippet."""
        if qualified_name not in state.def_index:
            return snippet_content

        func_info = state.def_index[qualified_name]

        # Check if the analyzer supports type reference extraction
        if not hasattr(lang_analyzer, "extract_type_references"):
            return snippet_content

        # Extract type references from the function node
        type_refs = lang_analyzer.extract_type_references(func_info.node)
        if not type_refs:
            return snippet_content

        # Find matching type definitions
        type_sources: list[str] = []
        for type_name in sorted(type_refs):
            if type_name in type_lookup:
                type_def = type_lookup[type_name]
                type_source = self._get_type_source(type_def, file_contents)
                if type_source:
                    type_sources.append(type_source)

        if not type_sources:
            return snippet_content

        # Prepend types to snippet
        return "\n\n".join(type_sources) + "\n\n" + snippet_content

    def _get_type_source(
        self, type_def: "TypeDefinition", file_contents: dict[Path, str]
    ) -> str | None:
        """Get the source code for a type definition."""
        file_content = self._get_file_content(type_def.file, file_contents)
        source_bytes = file_content.encode("utf-8")
        start_byte, end_byte = type_def.span
        type_source = source_bytes[start_byte:end_byte].decode("utf-8")
        return type_source.strip() if type_source.strip() else None



"""SQLAlchemy implementation of GitTagRepository."""

from collections.abc import Callable

from sqlalchemy.ext.asyncio import AsyncSession

from kodit.domain.entities.git import GitTag
from kodit.domain.protocols import GitTagRepository
from kodit.infrastructure.sqlalchemy import entities as db_entities
from kodit.infrastructure.sqlalchemy.query import FilterOperator, QueryBuilder
from kodit.infrastructure.sqlalchemy.repository import SqlAlchemyRepository


def create_git_tag_repository(
    session_factory: Callable[[], AsyncSession],
) -> GitTagRepository:
    """Create a git tag repository."""
    return SqlAlchemyGitTagRepository(session_factory=session_factory)


class SqlAlchemyGitTagRepository(
    SqlAlchemyRepository[GitTag, db_entities.GitTag], GitTagRepository
):
    """SQLAlchemy implementation of GitTagRepository."""

    def _get_id(self, entity: GitTag) -> tuple[int, str]:
        """Get the ID of a tag."""
        if entity.repo_id is None:
            raise ValueError("Repository ID is required")
        return (entity.repo_id, entity.name)

    @property
    def db_entity_type(self) -> type[db_entities.GitTag]:
        """The SQLAlchemy model type."""
        return db_entities.GitTag

    @staticmethod
    def to_domain(db_entity: db_entities.GitTag) -> GitTag:
        """Map a SQLAlchemy GitTag to a domain GitTag."""
        return GitTag(
            created_at=db_entity.created_at,
            updated_at=db_entity.updated_at,
            repo_id=db_entity.repo_id,
            name=db_entity.name,
            target_commit_sha=db_entity.target_commit_sha,
        )

    @staticmethod
    def to_db(domain_entity: GitTag) -> db_entities.GitTag:
        """Map a domain GitTag to a SQLAlchemy GitTag."""
        if domain_entity.repo_id is None:
            raise ValueError("Repository ID is required")
        return db_entities.GitTag(
            repo_id=domain_entity.repo_id,
            name=domain_entity.name,
            target_commit_sha=domain_entity.target_commit_sha,
        )

    async def get_by_name(self, tag_name: str, repo_id: int) -> GitTag:
        """Get a tag by name and repository ID."""
        query = (
            QueryBuilder()
            .filter("name", FilterOperator.EQ, tag_name)
            .filter("repo_id", FilterOperator.EQ, repo_id)
        )
        tags = await self.find(query)
        if not tags:
            raise ValueError(f"Tag {tag_name} not found in repo {repo_id}")
        return tags[0]

    async def get_by_repo_id(self, repo_id: int) -> list[GitTag]:
        """Get all tags for a repository."""
        return await self.find(
            QueryBuilder().filter("repo_id", FilterOperator.EQ, repo_id)
        )

    async def delete_by_repo_id(self, repo_id: int) -> None:
        """Delete all tags for a repository."""
        await self.delete_by_query(
            QueryBuilder().filter("repo_id", FilterOperator.EQ, repo_id)
        )



"""Tests for ASTAnalyzer."""

from datetime import UTC, datetime
from pathlib import Path

import pytest

from kodit.domain.entities.git import GitFile
from kodit.infrastructure.slicing.ast_analyzer import ASTAnalyzer


def test_ast_analyzer_initialization() -> None:
    """Test that ASTAnalyzer initializes correctly."""
    analyzer = ASTAnalyzer("python")
    assert analyzer.language == "python"
    assert analyzer.analyzer is not None


def test_ast_analyzer_unsupported_language() -> None:
    """Test that ASTAnalyzer raises error for unsupported language."""
    with pytest.raises(ValueError, match="Unsupported language"):
        ASTAnalyzer("invalid_lang")


def test_parse_files_with_python_file(tmp_path: Path) -> None:
    """Test parsing a simple Python file."""
    # Create a test Python file
    test_file = tmp_path / "test.py"
    test_file.write_text("def hello():\n    return 'world'\n")

    # Create GitFile entity
    git_file = GitFile(
        path=str(test_file),
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        blob_sha="test123",
        commit_sha="commit123",
        mime_type="text/x-python",
        size=100,
        extension=".py",
    )

    analyzer = ASTAnalyzer("python")
    parsed_files = analyzer.parse_files([git_file])

    assert len(parsed_files) == 1
    assert parsed_files[0].path == test_file
    assert parsed_files[0].tree is not None


def test_extract_definitions_from_python_file(tmp_path: Path) -> None:
    """Test extracting function definitions from Python file."""
    # Create a test Python file
    test_file = tmp_path / "test.py"
    test_file.write_text(
        """
def hello():
    return 'world'

def _private_func():
    return 'private'

def public_func():
    return 'public'
"""
    )

    git_file = GitFile(
        path=str(test_file),
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        blob_sha="test123",
        commit_sha="commit123",
        mime_type="text/x-python",
        size=100,
        extension=".py",
    )

    analyzer = ASTAnalyzer("python")
    parsed_files = analyzer.parse_files([git_file])
    functions, classes, types = analyzer.extract_definitions(parsed_files)

    # Should extract all functions including private when include_private=True
    assert len(functions) >= 3
    function_names = [f.simple_name for f in functions]
    assert "hello" in function_names
    assert "_private_func" in function_names
    assert "public_func" in function_names


def test_extract_definitions_filters_private(tmp_path: Path) -> None:
    """Test extracting only public functions."""
    test_file = tmp_path / "test.py"
    test_file.write_text(
        """
def hello():
    return 'world'

def _private_func():
    return 'private'
"""
    )

    git_file = GitFile(
        path=str(test_file),
        created_at=datetime(2024, 1, 1, tzinfo=UTC),
        blob_sha="test123",
        commit_sha="commit123",
        mime_type="text/x-python",
        size=100,
        extension=".py",
    )

    analyzer = ASTAnalyzer("python")
    parsed_files = analyzer.parse_files([git_file])
    functions, _, _ = analyzer.extract_definitions(parsed_files, include_private=False)

    function_names = [f.simple_name for f in functions]
    assert "hello" in function_names
    assert "_private_func" not in function_names



# ruff: noqa
"""add summary

Revision ID: 4552eb3f23ce
Revises: 9e53ea8bb3b0
Create Date: 2025-06-30 16:32:49.293087

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "4552eb3f23ce"
down_revision: Union[str, None] = "9e53ea8bb3b0"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("snippets", sa.Column("summary", sa.UnicodeText(), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("snippets", "summary")
    # ### end Alembic commands ###



"""Infrastructure implementation of ignore pattern provider."""

from pathlib import Path

import git
import pathspec

from kodit.infrastructure.git.git_utils import is_valid_clone_target


class GitIgnorePatternProvider:
    """Ignore pattern provider for git repositories."""

    def __init__(self, base_dir: Path) -> None:
        """Initialize the ignore pattern provider.

        Args:
            base_dir: The base directory to check for ignore patterns.

        Raises:
            ValueError: If the base directory is not a directory.

        """
        if not base_dir.is_dir():
            msg = f"Base directory is not a directory: {base_dir}"
            raise ValueError(msg)

        self.base_dir = base_dir

        # Check if the base_dir is a valid git repository
        self.git_repo = None
        if is_valid_clone_target(str(base_dir)):
            self.git_repo = git.Repo(base_dir)

    def should_ignore(self, path: Path) -> bool:
        """Check if a path should be ignored.

        Args:
            path: The path to check.

        Returns:
            True if the path should be ignored, False otherwise.

        """
        if path.is_dir():
            return False

        # Get the path relative to the base_dir
        relative_path = path.relative_to(self.base_dir)

        # If this file is _part_ of a .git directory, then it should be ignored
        if relative_path.as_posix().startswith(".git"):
            return True

        # If it is a git repository, then we need to check if the file is ignored
        if self.git_repo and len(self.git_repo.ignored(path)) > 0:
            return True

        # If the repo has a .noindex file
        noindex_path = Path(self.base_dir / ".noindex")
        if noindex_path.exists():
            with noindex_path.open() as f:
                patterns = [line.strip() for line in f if line.strip()]
                if patterns:
                    spec = pathspec.PathSpec.from_lines("gitwildmatch", patterns)
                    if spec.match_file(relative_path.as_posix()):
                        return True

        return False



"""Factory for creating embedding services with DDD architecture."""

from collections.abc import Callable

import structlog
from sqlalchemy.ext.asyncio import AsyncSession

from kodit.config import AppContext, Endpoint
from kodit.domain.services.embedding_service import (
    EmbeddingDomainService,
    EmbeddingProvider,
    VectorSearchRepository,
)
from kodit.infrastructure.embedding.embedding_providers.litellm_embedding_provider import (  # noqa: E501
    LiteLLMEmbeddingProvider,
)
from kodit.infrastructure.embedding.embedding_providers.local_embedding_provider import (  # noqa: E501
    CODE,
    LocalEmbeddingProvider,
)
from kodit.infrastructure.embedding.local_vector_search_repository import (
    LocalVectorSearchRepository,
)
from kodit.infrastructure.embedding.vectorchord_vector_search_repository import (
    TaskName,
    VectorChordVectorSearchRepository,
)
from kodit.infrastructure.sqlalchemy.embedding_repository import (
    create_embedding_repository,
)
from kodit.infrastructure.sqlalchemy.entities import EmbeddingType
from kodit.log import log_event


def _get_endpoint_configuration(app_context: AppContext) -> Endpoint | None:
    """Get the endpoint configuration for the embedding service."""
    return app_context.embedding_endpoint or None


def embedding_domain_service_factory(
    task_name: TaskName,
    app_context: AppContext,
    session_factory: Callable[[], AsyncSession],
) -> EmbeddingDomainService:
    """Create an embedding domain service."""
    structlog.get_logger(__name__)
    # Create embedding repository
    embedding_repository = create_embedding_repository(session_factory=session_factory)

    # Create embedding provider
    embedding_provider: EmbeddingProvider | None = None
    endpoint = _get_endpoint_configuration(app_context)

    if endpoint:
        log_event("kodit.embedding", {"provider": "litellm"})
        embedding_provider = LiteLLMEmbeddingProvider(endpoint=endpoint)
    else:
        log_event("kodit.embedding", {"provider": "local"})
        embedding_provider = LocalEmbeddingProvider(CODE)

    # Create vector search repository based on configuration
    vector_search_repository: VectorSearchRepository | None = None
    if app_context.default_search.provider == "vectorchord":
        log_event("kodit.database", {"provider": "vectorchord"})
        vector_search_repository = VectorChordVectorSearchRepository(
            session_factory=session_factory,
            task_name=task_name,
            embedding_provider=embedding_provider,
        )
    elif app_context.default_search.provider == "sqlite":
        log_event("kodit.database", {"provider": "sqlite"})
        if task_name == "code":
            embedding_type = EmbeddingType.CODE
        elif task_name == "text":
            embedding_type = EmbeddingType.TEXT
        else:
            raise ValueError(f"Invalid task name: {task_name}")

        vector_search_repository = LocalVectorSearchRepository(
            embedding_repository=embedding_repository,
            embedding_provider=embedding_provider,
            embedding_type=embedding_type,
        )
    else:
        msg = f"Invalid semantic search provider: {app_context.default_search.provider}"
        raise ValueError(msg)

    # Create and return domain service
    return EmbeddingDomainService(
        embedding_provider=embedding_provider,
        vector_search_repository=vector_search_repository,
    )



"""Application services for commit indexing operations."""

import structlog
from pydantic import AnyUrl

from kodit.application.handlers.registry import TaskHandlerRegistry
from kodit.application.services.queue_service import QueueService
from kodit.application.services.reporting import ProgressTracker
from kodit.domain.entities import Task
from kodit.domain.entities.git import GitRepo
from kodit.domain.factories.git_repo_factory import GitRepoFactory
from kodit.domain.protocols import GitRepoRepository, TaskStatusRepository
from kodit.domain.value_objects import (
    PrescribedOperations,
    QueuePriority,
    TaskOperation,
)
from kodit.infrastructure.sqlalchemy.query import TaskStatusQueryBuilder


class CommitIndexingApplicationService:
    """Application service for commit indexing operations."""

    def __init__(
        self,
        repo_repository: GitRepoRepository,
        task_status_repository: TaskStatusRepository,
        operation: ProgressTracker,
        queue: QueueService,
        handler_registry: TaskHandlerRegistry,
    ) -> None:
        """Initialize the commit indexing application service."""
        self.repo_repository = repo_repository
        self.task_status_repository = task_status_repository
        self.operation = operation
        self.queue = queue
        self.handler_registry = handler_registry
        self._log = structlog.get_logger(__name__)

    async def create_git_repository(self, remote_uri: AnyUrl) -> tuple[GitRepo, bool]:
        """Create a new Git repository or get existing one.

        Returns tuple of (repository, created) where created is True if new.
        """
        repo = GitRepoFactory.create_from_remote_uri(remote_uri)
        repo, created = await self.repo_repository.get_or_create(
            repo, "sanitized_remote_uri"
        )
        # Update remote_uri if credentials have changed
        if not created and str(repo.remote_uri) != str(remote_uri):
            repo.remote_uri = remote_uri
            repo = await self.repo_repository.save(repo)

        # Clear existing task statuses immediately for fresh progress tracking
        if repo.id:
            await self.task_status_repository.delete_by_query(
                TaskStatusQueryBuilder().for_repository(repo.id)
            )

        await self.queue.enqueue_tasks(
            tasks=PrescribedOperations.CREATE_NEW_REPOSITORY,
            base_priority=QueuePriority.USER_INITIATED,
            payload={"repository_id": repo.id},
        )
        return repo, created

    async def delete_git_repository(self, repo_id: int) -> bool:
        """Delete a Git repository by ID."""
        repo = await self.repo_repository.get(repo_id)
        if not repo:
            return False

        # Use the handler to delete the repository
        handler = self.handler_registry.handler(TaskOperation.DELETE_REPOSITORY)
        await handler.execute({"repository_id": repo_id})
        return True

    async def rescan_commit(self, repo_id: int, commit_sha: str) -> None:
        """Rescan a commit by deleting artifacts and re-triggering indexing."""
        # Clear existing task statuses immediately for fresh progress tracking
        await self.task_status_repository.delete_by_query(
            TaskStatusQueryBuilder().for_repository(repo_id)
        )

        handler = self.handler_registry.handler(TaskOperation.RESCAN_COMMIT)
        await handler.execute({
            "repository_id": repo_id,
            "commit_sha": commit_sha,
        })

    async def run_task(self, task: Task) -> None:
        """Run a task by delegating to the appropriate handler."""
        handler = self.handler_registry.handler(task.type)
        await handler.execute(task.payload)



"""Development enrichment domain entity."""

from abc import ABC
from dataclasses import dataclass

from kodit.domain.enrichments.enrichment import CommitEnrichment

ENRICHMENT_TYPE_DEVELOPMENT = "development"


@dataclass(frozen=True)
class DevelopmentEnrichment(CommitEnrichment, ABC):
    """Enrichment containing development discovery for a commit."""

    @property
    def type(self) -> str:
        """Return the enrichment type."""
        return ENRICHMENT_TYPE_DEVELOPMENT



"""Tracking configuration JSON-API schemas."""

from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field

from kodit.domain.entities.git import GitRepo, TrackingConfig, TrackingType


class TrackingMode(StrEnum):
    """Tracking mode for repository configuration."""

    BRANCH = "branch"
    TAG = "tag"


class TrackingConfigAttributes(BaseModel):
    """Tracking configuration attributes following JSON-API spec."""

    mode: TrackingMode = Field(
        ...,
        description="'branch' tracks a specific branch, 'tag' tracks the latest tag",
    )
    value: str | None = Field(
        None,
        description="Branch name when mode is 'branch'. Not used for 'tag' mode.",
    )

    @staticmethod
    def from_tracking_config(
        config: TrackingConfig | None,
    ) -> "TrackingConfigAttributes":
        """Create tracking config attributes from domain TrackingConfig."""
        if not config:
            return TrackingConfigAttributes(mode=TrackingMode.BRANCH, value="main")

        if config.type == TrackingType.TAG:
            return TrackingConfigAttributes(mode=TrackingMode.TAG, value=None)

        return TrackingConfigAttributes(mode=TrackingMode.BRANCH, value=config.name)

    def to_domain(self) -> TrackingConfig:
        """Convert to domain TrackingConfig."""
        if self.mode == TrackingMode.TAG:
            return TrackingConfig(type=TrackingType.TAG, name="")

        return TrackingConfig(type=TrackingType.BRANCH, name=self.value or "main")


class TrackingConfigData(BaseModel):
    """Tracking configuration data following JSON-API spec."""

    type: Literal["tracking-config"] = "tracking-config"
    attributes: TrackingConfigAttributes

    @staticmethod
    def from_git_repo(repo: GitRepo) -> "TrackingConfigData":
        """Create tracking config data from a Git repository."""
        return TrackingConfigData(
            attributes=TrackingConfigAttributes.from_tracking_config(repo.tracking_config),
        )


class TrackingConfigResponse(BaseModel):
    """Tracking configuration response following JSON-API spec."""

    data: TrackingConfigData


class TrackingConfigUpdateAttributes(BaseModel):
    """Tracking configuration update attributes."""

    mode: TrackingMode = Field(
        ...,
        description="'branch' tracks a specific branch, 'tag' tracks the latest tag",
    )
    value: str | None = Field(
        None,
        description="Branch name when mode is 'branch'. Not used for 'tag' mode.",
    )

    def to_domain(self) -> TrackingConfig:
        """Convert to domain TrackingConfig."""
        if self.mode == TrackingMode.TAG:
            return TrackingConfig(type=TrackingType.TAG, name="")

        return TrackingConfig(type=TrackingType.BRANCH, name=self.value or "main")


class TrackingConfigUpdateData(BaseModel):
    """Tracking configuration update data."""

    type: Literal["tracking-config"] = "tracking-config"
    attributes: TrackingConfigUpdateAttributes


class TrackingConfigUpdateRequest(BaseModel):
    """Tracking configuration update request."""

    data: TrackingConfigUpdateData



"""Tests for the Slicer."""

from datetime import UTC, datetime
from pathlib import Path

import pytest

from kodit.domain.entities.git import GitFile
from kodit.infrastructure.slicing.slicer import Slicer


class TestSlicer:
    """Test the Slicer functionality."""

    @pytest.mark.parametrize(
        ("language", "extension"),
        [
            ("c", ".c"),
            ("cpp", ".cpp"),
            ("csharp", ".cs"),
            ("go", ".go"),
            ("java", ".java"),
            ("javascript", ".js"),
            ("python", ".py"),
            ("rust", ".rs"),
            ("tsx", ".tsx"),
        ],
    )
    def test_extract_snippets_from_language(
        self, language: str, extension: str
    ) -> None:
        """Test extracting snippets from supported languages."""
        data_dir = Path(__file__).parent / "data" / language
        files = [f for f in data_dir.glob(f"*{extension}") if f.is_file()]

        git_files = [
            GitFile(
                created_at=datetime.now(tz=UTC),
                blob_sha=f"sha_{f.name}",
                commit_sha="test_commit",
                path=str(f),
                mime_type="text/plain",
                size=f.stat().st_size,
                extension=extension,
            )
            for f in files
        ]

        slicer = Slicer()
        snippets = slicer.extract_snippets_from_git_files(git_files, language)

        assert len(snippets) > 0
        assert all(snippet.extension == language for snippet in snippets)



"""Domain services for Git repository scanning and cloning operations."""

import shutil
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import structlog
from pydantic import AnyUrl

if TYPE_CHECKING:
    from kodit.domain.protocols import GitAdapter

from kodit.domain.entities import WorkingCopy
from kodit.domain.entities.git import (
    GitCommit,
    GitFile,
    GitRepo,
    TrackingConfig,
    TrackingType,
)


@dataclass(frozen=True)
class RepositoryInfo:
    """Immutable repository information needed for GitRepo construction."""

    remote_uri: AnyUrl
    sanitized_remote_uri: AnyUrl
    cloned_path: Path


class GitRepositoryScanner:
    """Pure scanner that extracts data without mutation."""

    def __init__(self, git_adapter: "GitAdapter") -> None:
        """Initialize the Git repository scanner."""
        self._log = structlog.getLogger(__name__)
        self.git_adapter = git_adapter

    async def scan_commit(
        self,
        cloned_path: Path,
        commit_sha: str,
        repo_id: int,
    ) -> tuple[GitCommit, list[GitFile]]:
        """Scan a specific commit and return commit with its files.

        Args:
            cloned_path: Path to the cloned repository
            commit_sha: SHA of the commit to scan
            repo_id: Repository ID

        Returns:
            Tuple of (commit, files) for the scanned commit

        """
        self._log.info(f"Scanning commit {commit_sha[:8]} at: {cloned_path}")

        # Get commit details
        commit_data = await self.git_adapter.get_commit_details(cloned_path, commit_sha)

        # Create GitCommit entity
        git_commit = self._create_lightweight_git_commit(
            commit_data, datetime.now(UTC), repo_id
        )
        if not git_commit:
            raise ValueError(f"Failed to create commit object for {commit_sha}")

        # Get files for this commit
        files_data = await self.git_adapter.get_commit_files(cloned_path, commit_sha)
        files = self._create_git_files(cloned_path, files_data, commit_sha)

        self._log.info(f"Scanned commit {commit_sha[:8]}: found {len(files)} files")

        return git_commit, files

    def _format_author_from_data(self, commit_data: dict[str, Any]) -> str:
        """Format author string from commit data."""
        author_name = commit_data.get("author_name", "")
        author_email = commit_data.get("author_email", "")
        if author_name and author_email:
            return f"{author_name} <{author_email}>"
        return author_name or "Unknown"

    def _create_lightweight_git_commit(
        self, commit_data: dict[str, Any], created_at: datetime, repo_id: int | None
    ) -> GitCommit | None:
        """Create a GitCommit without expensive file data fetching."""
        try:
            commit_sha = commit_data["sha"]
            author = self._format_author_from_data(commit_data)

            # Create commit with empty files list for now
            # Files will be loaded lazily when actually needed (e.g., during indexing)
            return GitCommit(
                created_at=created_at,
                commit_sha=commit_sha,
                repo_id=repo_id or 0,  # Use 0 as default if not provided
                date=commit_data["date"],
                message=commit_data["message"],
                parent_commit_sha=commit_data["parent_sha"],
                author=author,
            )
        except Exception as e:  # noqa: BLE001
            self._log.warning(f"Failed to create commit {commit_data.get('sha')}: {e}")
            return None

    def _create_git_files(
        self, cloned_path: Path, files_data: list[dict], commit_sha: str
    ) -> list[GitFile]:
        """Create GitFile entities from files data."""
        # Cache expensive path operations
        cloned_path_str = str(cloned_path)
        current_time = datetime.now(UTC)

        result = []
        for f in files_data:
            # Avoid expensive Path operations by doing string concatenation
            file_path = f["path"]
            full_path = f"{cloned_path_str}/{file_path}"

            result.append(
                GitFile(
                    blob_sha=f["blob_sha"],
                    commit_sha=commit_sha,
                    path=full_path,
                    mime_type=f.get("mime_type", "application/octet-stream"),
                    size=f["size"],
                    extension=GitFile.extension_from_path(file_path),
                    created_at=f.get("created_at", current_time),
                )
            )
        return result

    async def process_files_for_commits_batch(
        self, cloned_path: Path, commit_shas: list[str]
    ) -> list[GitFile]:
        """Process files for a batch of commits.

        This allows the application service to process files in batches
        to avoid loading millions of files into memory at once.

        CRITICAL: Reuses a single Repo object to avoid creating 32K+ Repo instances
        which would consume massive memory (1-2 MB each).
        """
        from git import Repo

        # Open repo once and reuse for all commits in this batch
        repo = Repo(cloned_path)
        files = []

        try:
            for commit_sha in commit_shas:
                files_data = await self.git_adapter.get_commit_files(
                    cloned_path, commit_sha
                )
                files.extend(
                    self._create_git_files(cloned_path, files_data, commit_sha)
                )
        finally:
            # Explicitly close the repo to free resources
            repo.close()

        return files


class RepositoryCloner:
    """Pure service for cloning repositories."""

    def __init__(self, git_adapter: "GitAdapter", clone_dir: Path) -> None:
        """Initialize the repository cloner."""
        self._log = structlog.getLogger(__name__)
        self.git_adapter = git_adapter
        self.clone_dir = clone_dir

    def clone_path_from_uri(self, uri: AnyUrl) -> Path:
        """Get the clone path for a Git repository."""
        sanitized_uri = WorkingCopy.sanitize_git_url(str(uri))
        dir_name = GitRepo.create_id(sanitized_uri)
        return self.clone_dir / dir_name

    async def clone_repository(self, remote_uri: AnyUrl, clone_path: Path) -> Path:
        """Clone repository and return repository info."""
        try:
            await self.git_adapter.clone_repository(str(remote_uri), clone_path)
        except Exception:
            shutil.rmtree(clone_path)
            raise

        return clone_path

    async def update_repository(self, repository: GitRepo) -> None:
        """Update repository based on tracking configuration."""
        if not repository.cloned_path:
            raise ValueError(f"Repository {repository.id} has never been cloned")

        if repository.cloned_path and not repository.cloned_path.exists():
            # Re-clone this repository if the setting exists but the dir does not
            await self.clone_repository(repository.remote_uri, repository.cloned_path)

        if not repository.tracking_config:
            self._log.debug(
                "Repository has no tracking config yet", repo_id=repository.id
            )
            return

        if repository.tracking_config.type == TrackingType.BRANCH:
            await self.git_adapter.fetch_repository(repository.cloned_path)
            try:
                await self.git_adapter.checkout_branch(
                    repository.cloned_path, repository.tracking_config.name
                )
            except Exception as e:  # noqa: BLE001
                # Branch might not exist - re-detect default branch
                self._log.warning(
                    "Checkout failed, re-detecting default branch",
                    branch=repository.tracking_config.name,
                    error=str(e),
                )
                new_default = await self.git_adapter.get_default_branch(
                    repository.cloned_path
                )
                repository.tracking_config = TrackingConfig(
                    type=TrackingType.BRANCH, name=new_default
                )
                await self.git_adapter.checkout_branch(
                    repository.cloned_path, new_default
                )
            await self.git_adapter.pull_repository(repository.cloned_path)
        elif repository.tracking_config.type == TrackingType.TAG:
            # Fetch all tags and get the most recent
            await self.git_adapter.fetch_repository(repository.cloned_path)
            tags = await self.git_adapter.get_all_tags(repository.cloned_path)

            latest_tag = max(tags, key=lambda t: t["target_commit_sha"])
            commit_sha = latest_tag["target_commit_sha"]
            await self.git_adapter.checkout_commit(repository.cloned_path, commit_sha)
        else:
            raise ValueError(
                f"Invalid tracking type: {repository.tracking_config.type}"
            )



"""Docker Compose detector for physical architecture discovery."""

import contextlib
import re
from pathlib import Path

import yaml


class DockerComposeDetector:
    """Detects physical components from Docker Compose files and generates narrative observations."""  # noqa: E501

    # Regex pattern to detect communication addresses in environment variables
    # Matches complete URLs with hostnames:
    # - Simple URLs: http://api:8080, redis://cache:6379
    # - Connection strings with auth: postgresql://user:pass@db:5432/dbname
    # - Connection strings with asyncpg: postgresql+asyncpg://user:pass@db:5432
    # Note: This captures the hostname portion, avoiding false matches in
    # passwords or other parts of the URL
    COMMUNICATION_PATTERN = re.compile(
        r"(?:"
        # Protocol-based URLs with optional auth (user:pass@)
        r"(?:https?|tcp|grpc|ws|wss|amqp|kafka|redis|memcached|"
        r"postgres(?:ql)?(?:\+\w+)?|mysql|mongodb)://"
        r"(?:[^@/]+@)?"  # Optional user:pass@ (non-capturing, skip it)
        r"([\w\-\.]+(?::\d+)?)"  # Capture hostname:port after @ or ://
        r")",
        re.IGNORECASE,
    )

    async def analyze(self, repo_path: Path) -> tuple[list[str], list[str], list[str]]:
        """Generate narrative observations from Docker Compose analysis."""
        component_notes: list[str] = []
        connection_notes: list[str] = []
        infrastructure_notes: list[str] = []

        # Find all docker-compose files
        yml_files = list(repo_path.glob("docker-compose*.yml"))
        yaml_files = list(repo_path.glob("docker-compose*.yaml"))
        compose_files = yml_files + yaml_files

        if not compose_files:
            return ([], [], [])

        # Analyze each compose file
        for compose_file in compose_files:
            try:
                with compose_file.open(encoding="utf-8") as f:
                    compose_data = yaml.safe_load(f)

                if not compose_data or "services" not in compose_data:
                    continue

                self._analyze_compose_file(
                    compose_file,
                    compose_data,
                    component_notes,
                    connection_notes,
                    infrastructure_notes,
                )

            except (yaml.YAMLError, OSError, KeyError):
                infrastructure_notes.append(
                    f"Unable to parse Docker Compose file at {compose_file}. "
                    "File may be malformed or inaccessible."
                )

        return (component_notes, connection_notes, infrastructure_notes)

    def _analyze_compose_file(
        self,
        compose_file: Path,
        compose_data: dict,
        component_notes: list[str],
        connection_notes: list[str],
        infrastructure_notes: list[str],
    ) -> None:
        """Analyze a single Docker Compose file and generate observations."""
        services = compose_data.get("services", {})

        # High-level infrastructure observation
        infrastructure_notes.append(
            f"Found Docker Compose configuration at {compose_file.name} defining "
            f"{len(services)} services. This suggests a containerized application "
            f"architecture with orchestrated service dependencies."
        )

        # Analyze each service
        for service_name, service_config in services.items():
            self._analyze_service(
                service_name,
                service_config,
                component_notes,
                connection_notes,
            )

        # Analyze service dependencies
        self._analyze_service_dependencies(services, connection_notes)

        # Check for additional Docker Compose features
        self._analyze_compose_features(compose_data, infrastructure_notes)

    def _analyze_service(
        self,
        service_name: str,
        service_config: dict,
        component_notes: list[str],
        _connection_notes: list[str],
    ) -> None:
        """Generate narrative observations for a single service."""
        # Extract key configuration details
        image = service_config.get("image", "")
        build = service_config.get("build", "")
        ports = self._extract_ports(service_config)

        component_observation = (
            f"Found '{service_name}' service in Docker Compose configuration."
        )

        # Add deployment details
        if image:
            component_observation += f" Service uses '{image}' Docker image"
            if ":" in image:
                tag = image.split(":")[-1]
                component_observation += f" with tag '{tag}'"
            component_observation += "."
        elif build:
            component_observation += f" Service builds from local source at '{build}'."

        # Add port information
        if ports:
            port_list = ", ".join(str(p) for p in ports)
            component_observation += f" Exposes ports {port_list}"
            protocol_info = self._infer_protocol_description(ports)
            if protocol_info:
                component_observation += f" suggesting {protocol_info}"
            component_observation += "."

        component_notes.append(component_observation)

    def _analyze_service_dependencies(  # noqa: PLR0912, C901
        self, services: dict, connection_notes: list[str]
    ) -> None:
        """Analyze dependencies between services."""
        for service_name, service_config in services.items():
            depends_on = service_config.get("depends_on", [])

            if isinstance(depends_on, dict):
                dependencies = list(depends_on.keys())
                condition_info = []
                for dep, condition in depends_on.items():
                    if isinstance(condition, dict) and "condition" in condition:
                        condition_info.append(f"{dep} ({condition['condition']})")

                if condition_info:
                    connection_notes.append(
                        f"Service '{service_name}' has conditional dependencies on "
                        f"{', '.join(condition_info)}, indicating sophisticated "
                        "startup orchestration with health checks."
                    )
                else:
                    dependencies = list(depends_on.keys())
            elif isinstance(depends_on, list):
                dependencies = depends_on
            else:
                continue

            if dependencies:
                dep_list = "', '".join(dependencies)
                connection_notes.append(
                    f"Docker Compose 'depends_on' configuration shows '{service_name}' "
                    f"requires '{dep_list}' to start first, indicating service startup "
                    "dependency and likely runtime communication pattern."
                )

        # Check for communication patterns in environment variables
        # and command arguments
        service_names = {name for name, _ in services.items()}
        # Track which connections we've already recorded to avoid duplicates
        recorded_connections: set[tuple[str, str]] = set()

        for service_name, service_config in services.items():
            # Check environment variables
            env = service_config.get("environment", [])
            if isinstance(env, list):
                for var in env:
                    self._check_communication_pattern(
                        var,
                        service_name,
                        service_names,
                        "environment variable",
                        connection_notes,
                        recorded_connections,
                    )
            elif isinstance(env, dict):
                for value in env.values():
                    if isinstance(value, str):
                        self._check_communication_pattern(
                            value,
                            service_name,
                            service_names,
                            "environment variable",
                            connection_notes,
                            recorded_connections,
                        )

            # Check command arguments
            args = service_config.get("command", [])
            if isinstance(args, list):
                for arg in args:
                    if isinstance(arg, str):
                        self._check_communication_pattern(
                            arg,
                            service_name,
                            service_names,
                            "command argument",
                            connection_notes,
                            recorded_connections,
                        )
            elif isinstance(args, str):
                self._check_communication_pattern(
                    args,
                    service_name,
                    service_names,
                    "command argument",
                    connection_notes,
                    recorded_connections,
                )

    def _check_communication_pattern(  # noqa: PLR0913
        self,
        text: str,
        service_name: str,
        service_names: set[str],
        source_type: str,
        connection_notes: list[str],
        recorded_connections: set[tuple[str, str]],
    ) -> None:
        """Check if text contains communication patterns referencing other services."""
        # Find all matches and extract hostnames from captured groups
        matches = self.COMMUNICATION_PATTERN.finditer(text)
        hostnames = set()

        for match in matches:
            # Group 1 contains the hostname
            if match.group(1):
                # Extract just the hostname (without port)
                hostname = match.group(1).split(":")[0]
                hostnames.add(hostname)

        if not hostnames:
            return

        # Check if any extracted hostname matches a service name
        for target_service in service_names:
            if target_service == service_name:
                continue

            # Check if the target service is in the extracted hostnames
            if target_service in hostnames:
                connection_key = (service_name, target_service)
                if connection_key not in recorded_connections:
                    connection_notes.append(
                        f"'{service_name}' has a communication address referencing "
                        f"'{target_service}' in its {source_type}, indicating a "
                        "direct runtime dependency."
                    )
                    recorded_connections.add(connection_key)
                    break

    def _analyze_compose_features(
        self, compose_data: dict, infrastructure_notes: list[str]
    ) -> None:
        """Analyze additional Docker Compose features."""
        # Check for networks
        networks = compose_data.get("networks", {})
        if networks:
            infrastructure_notes.append(
                f"Docker Compose defines {len(networks)} custom networks, "
                "indicating network segmentation and controlled service communication."
            )

    def _extract_ports(self, service_config: dict) -> list[int]:
        """Extract port numbers from service configuration."""
        ports = []

        # Extract from 'ports' section
        port_specs = service_config.get("ports", [])
        for port_spec in port_specs:
            if isinstance(port_spec, str):
                if ":" in port_spec:
                    external_port = port_spec.split(":")[0]
                    with contextlib.suppress(ValueError):
                        ports.append(int(external_port))
                else:
                    with contextlib.suppress(ValueError):
                        ports.append(int(port_spec))
            elif isinstance(port_spec, int):
                ports.append(port_spec)

        # Extract from 'expose' section
        expose_specs = service_config.get("expose", [])
        for expose_spec in expose_specs:
            with contextlib.suppress(ValueError, TypeError):
                ports.append(int(expose_spec))

        return sorted(set(ports))

    def _infer_protocol_description(self, ports: list[int]) -> str:
        """Infer protocol information from ports and return descriptive text."""
        protocols = []

        # HTTP ports
        http_ports = {80, 8080, 3000, 4200, 5000, 8000, 8443, 443}
        if any(port in http_ports for port in ports):
            protocols.append("HTTP/HTTPS web traffic")

        # gRPC ports
        grpc_ports = {9090, 50051}
        if any(port in grpc_ports for port in ports):
            protocols.append("gRPC API communication")

        # Cache/Redis ports
        if 6379 in ports:
            protocols.append("cache service")

        # Database ports (excluding Redis which is handled above)
        db_ports = {5432, 3306, 27017}
        if any(port in db_ports for port in ports):
            protocols.append("database service")

        if protocols:
            return " and ".join(protocols)
        if ports:
            return "TCP-based service communication"
        return ""



"""MCP server for kodit."""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import structlog
from fastmcp import Context, FastMCP
from pydantic import Field

from kodit._version import version
from kodit.application.factories.server_factory import ServerFactory
from kodit.application.services.code_search_application_service import MultiSearchResult
from kodit.config import AppContext
from kodit.database import Database
from kodit.domain.value_objects import (
    MultiSearchRequest,
    SnippetSearchFilters,
)
from kodit.infrastructure.sqlalchemy.query import QueryBuilder

# Global database connection for MCP server
_mcp_db: Database | None = None
_mcp_server_factory: ServerFactory | None = None


@dataclass
class MCPContext:
    """Context for the MCP server."""

    server_factory: ServerFactory


@asynccontextmanager
async def mcp_lifespan(_: FastMCP) -> AsyncIterator[MCPContext]:
    """Lifespan context manager for the MCP server.

    This is called for each request. The MCP server is designed to work with both
    the CLI and the FastAPI server. Therefore, we must carefully reconstruct the
    application context. uvicorn does not pass through CLI args, so we must rely on
    parsing env vars set in the CLI.

    This lifespan is recreated for each request. See:
    https://github.com/jlowin/fastmcp/issues/166

    Since they don't provide a good way to handle global state, we must use a
    global variable to store the database connection.
    """
    global _mcp_server_factory  # noqa: PLW0603
    if _mcp_server_factory is None:
        app_context = AppContext()
        db = await app_context.get_db()
        _mcp_server_factory = ServerFactory(app_context, db.session_factory)
    yield MCPContext(_mcp_server_factory)


def create_mcp_server(name: str, instructions: str | None = None) -> FastMCP:
    """Create a FastMCP server with common configuration."""
    return FastMCP(
        name,
        lifespan=mcp_lifespan,
        instructions=instructions,
    )


def _format_enrichments(enrichments: list) -> str:
    """Format enrichments in a simple, LLM-friendly format.

    Just returns the content from each enrichment, separated by double newlines.
    """
    if not enrichments:
        return ""

    contents = [e.content for e in enrichments if e.content]
    return "\n\n".join(contents)


def register_mcp_tools(mcp_server: FastMCP) -> None:  # noqa: C901, PLR0915
    """Register MCP tools on the provided FastMCP instance."""

    @mcp_server.tool()
    async def search(  # noqa: PLR0913
        ctx: Context,
        user_intent: Annotated[
            str,
            Field(
                description="Think about what the user wants to achieve. Describe the "
                "user's intent in one sentence."
            ),
        ],
        related_file_paths: Annotated[
            list[Path],
            Field(
                description=(
                    "A list of absolute paths to files that are relevant to the "
                    "user's intent."
                )
            ),
        ],
        related_file_contents: Annotated[
            list[str],
            Field(
                description=(
                    "A list of the contents of the files that are relevant to the "
                    "user's intent."
                )
            ),
        ],
        keywords: Annotated[
            list[str],
            Field(
                description=(
                    "A list of keywords that are relevant to the desired outcome."
                )
            ),
        ],
        language: Annotated[
            str | None,
            Field(
                description="Filter by language (e.g., 'python', 'go', 'javascript')."
            ),
        ] = None,
        author: Annotated[
            str | None,
            Field(description=("Filter to search for snippets by a specific author.")),
        ] = None,
        created_after: Annotated[
            str | None,
            Field(
                description=(
                    "Filter for snippets created after this date "
                    "(ISO format: YYYY-MM-DD)."
                )
            ),
        ] = None,
        created_before: Annotated[
            str | None,
            Field(
                description=(
                    "Filter for snippets created before this date "
                    "(ISO format: YYYY-MM-DD)."
                )
            ),
        ] = None,
        source_repo: Annotated[
            str | None,
            Field(
                description=(
                    "Filter results by project source repository (e.g., "
                    "github.com/example/repo)"
                )
            ),
        ] = None,
        enrichment_subtypes: Annotated[
            list[str] | None,
            Field(
                description=(
                    "Filter by enrichment subtypes. Use ['example'] for full "
                    "example files and documentation code blocks, ['snippet'] for "
                    "AST-extracted code snippets. Leave empty to search both."
                )
            ),
        ] = None,
    ) -> str:
        """Search for relevant code snippets and examples across repositories.

        Use this when you need to find specific code patterns, implementations, or
        examples that match the user's requirements. This searches through actual
        code snippets indexed from repositories.
        """
        # This docstring is used by the AI assistant to decide when to call the tool.
        # If you want to update it, please make sure you thoroughly test the
        # assistant's response to the updated tool call.

        log = structlog.get_logger(__name__)

        log.debug(
            "Searching for relevant snippets",
            user_intent=user_intent,
            keywords=keywords,
            file_count=len(related_file_paths),
            file_paths=related_file_paths,
            file_contents=related_file_contents,
        )

        mcp_context: MCPContext = ctx.request_context.lifespan_context

        # Validate source_repo if provided
        if source_repo:
            repo_query_service = mcp_context.server_factory.repository_query_service()
            repo_id = await repo_query_service.find_repo_by_url(source_repo)
            if not repo_id:
                raise ValueError(f"Repository not found: {source_repo}")

        # Use the unified application service
        service = mcp_context.server_factory.code_search_application_service()

        log.debug("Searching for snippets")

        # Create filters if any filter parameters are provided
        filters = SnippetSearchFilters.from_cli_params(
            language=language,
            author=author,
            created_after=created_after,
            created_before=created_before,
            source_repo=source_repo,
            enrichment_subtypes=enrichment_subtypes,
        )

        search_request = MultiSearchRequest(
            keywords=keywords,
            code_query="\n".join(related_file_contents),
            text_query=user_intent,
            filters=filters,
        )

        log.debug("Searching for snippets")
        snippets = await service.search(request=search_request)

        log.debug("Fusing output")
        output = MultiSearchResult.to_markdown(results=snippets)

        log.debug("Output", output=output)
        return output

    @mcp_server.tool()
    async def get_version() -> str:
        """Get the version of the kodit project."""
        return version

    @mcp_server.tool()
    async def list_repositories(ctx: Context) -> str:
        """List all repositories available in the system.

        Call this first to discover which repositories you can query for documentation,
        schemas, cookbooks, and other enrichments. The returned repository URLs can be
        used with other tools like get_architecture_docs(), get_api_docs(), etc.

        Returns a list of repositories with their tracking information and latest
        commit SHA.
        """
        log = structlog.get_logger(__name__)
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_repository = mcp_context.server_factory.repo_repository()
        repo_query_service = mcp_context.server_factory.repository_query_service()

        repos = await repo_repository.find(QueryBuilder())

        if not repos:
            return "No repositories found."

        lines = ["Available repositories:"]
        for repo in repos:
            # Base repository info
            repo_info = f"- {repo.sanitized_remote_uri}"

            # Add tracking information if available
            if repo.tracking_config:
                tracking_type = repo.tracking_config.type
                tracking_name = repo.tracking_config.name
                repo_info += f" (tracking {tracking_type}: {tracking_name})"

                # Get the latest commit SHA for the tracked thing
                try:
                    if repo.id:
                        latest_commit = await repo_query_service.find_latest_commit(
                            repo_id=repo.id,
                            max_commits_to_check=1,
                        )
                        if latest_commit:
                            repo_info += f" [latest: {latest_commit[:8]}]"
                except ValueError as e:
                    # Log if we can't get the commit
                    log.debug(
                        "Could not get latest commit for repository",
                        repo_id=repo.id,
                        error=str(e),
                    )

            lines.append(repo_info)

        return "\n".join(lines)

    @mcp_server.tool()
    async def get_architecture_docs(
        ctx: Context,
        repo_url: Annotated[
            str,
            Field(description="The repository URL (e.g., github.com/user/repo)"),
        ],
        commit_sha: Annotated[
            str | None,
            Field(
                description=(
                    "Optional commit SHA. If not provided, uses most recent "
                    "commit with architecture docs"
                )
            ),
        ] = None,
    ) -> str:
        """Get high-level architecture documentation for a repository.

        Use this to understand:
        - Overall project structure and organization
        - Key architectural patterns and design decisions
        - Component relationships and boundaries
        - Module organization and responsibilities

        Prefer this over search() when you need to understand the big picture before
        diving into specific code examples.

        **Important**: If you don't know which repositories are available, call
        list_repositories() first to see the available options.
        """
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_query_service = mcp_context.server_factory.repository_query_service()

        # Find the repository by URL
        repo_id = await repo_query_service.find_repo_by_url(repo_url)
        if not repo_id:
            raise ValueError(f"Repository not found: {repo_url}")

        if not commit_sha:
            commit_sha = await repo_query_service.find_latest_commit(
                repo_id=repo_id,
                max_commits_to_check=100,
            )
            if not commit_sha:
                msg = (
                    f"No commits with architecture docs found "
                    f"for repository: {repo_url}"
                )
                raise ValueError(msg)

        enrichment_service = mcp_context.server_factory.enrichment_query_service()
        enrichments = await enrichment_service.get_architecture_docs_for_commit(
            commit_sha
        )
        return _format_enrichments(enrichments)

    @mcp_server.tool()
    async def get_api_docs(
        ctx: Context,
        repo_url: Annotated[
            str,
            Field(description="The repository URL (e.g., github.com/user/repo)"),
        ],
        commit_sha: Annotated[
            str | None,
            Field(
                description=(
                    "Optional commit SHA. If not provided, uses most recent "
                    "commit with API docs"
                )
            ),
        ] = None,
    ) -> str:
        """Get API documentation showing public interfaces and contracts.

        Use this to understand:
        - Public APIs and their usage patterns
        - Function signatures and parameters
        - Expected inputs and outputs
        - Interface contracts and guarantees

        Prefer this over search() when you need to understand how to interact with
        existing components rather than finding implementation examples.

        **Important**: If you don't know which repositories are available, call
        list_repositories() first to see the available options.
        """
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_query_service = mcp_context.server_factory.repository_query_service()

        # Find the repository by URL
        repo_id = await repo_query_service.find_repo_by_url(repo_url)
        if not repo_id:
            raise ValueError(f"Repository not found: {repo_url}")

        if not commit_sha:
            commit_sha = await repo_query_service.find_latest_commit(
                repo_id=repo_id,
                max_commits_to_check=100,
            )
            if not commit_sha:
                raise ValueError(
                    f"No commits with API docs found for repository: {repo_url}"
                )

        enrichment_service = mcp_context.server_factory.enrichment_query_service()
        enrichments = await enrichment_service.get_api_docs_for_commit(commit_sha)
        return _format_enrichments(enrichments)

    @mcp_server.tool()
    async def get_commit_description(
        ctx: Context,
        repo_url: Annotated[
            str,
            Field(description="The repository URL (e.g., github.com/user/repo)"),
        ],
        commit_sha: Annotated[
            str | None,
            Field(
                description=(
                    "Optional commit SHA. If not provided, uses most recent "
                    "commit with description"
                )
            ),
        ] = None,
    ) -> str:
        """Get human-readable descriptions of recent changes and their rationale.

        Use this to understand:
        - What recently changed in the codebase
        - Why changes were made
        - Context around recent development decisions
        - Evolution of the codebase over time

        Prefer this when you need to understand recent development context or when
        working with actively changing code.

        **Important**: If you don't know which repositories are available, call
        list_repositories() first to see the available options.
        """
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_query_service = mcp_context.server_factory.repository_query_service()

        # Find the repository by URL
        repo_id = await repo_query_service.find_repo_by_url(repo_url)
        if not repo_id:
            raise ValueError(f"Repository not found: {repo_url}")

        if not commit_sha:
            commit_sha = await repo_query_service.find_latest_commit(
                repo_id=repo_id,
                max_commits_to_check=100,
            )
            if not commit_sha:
                msg = (
                    f"No commits with commit description found "
                    f"for repository: {repo_url}"
                )
                raise ValueError(msg)

        enrichment_service = mcp_context.server_factory.enrichment_query_service()
        enrichments = await enrichment_service.get_commit_description_for_commit(
            commit_sha
        )
        return _format_enrichments(enrichments)

    @mcp_server.tool()
    async def get_database_schema(
        ctx: Context,
        repo_url: Annotated[
            str,
            Field(description="The repository URL (e.g., github.com/user/repo)"),
        ],
        commit_sha: Annotated[
            str | None,
            Field(
                description=(
                    "Optional commit SHA. If not provided, uses most recent "
                    "commit with database schema"
                )
            ),
        ] = None,
    ) -> str:
        """Get database schema documentation.

        Use this to understand:
        - Database table structures
        - Column types and constraints
        - Relationships between entities
        - ORM model definitions

        Prefer this over search() when you need to understand data models or write
        database-related code.

        **Important**: If you don't know which repositories are available, call
        list_repositories() first to see the available options.
        """
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_query_service = mcp_context.server_factory.repository_query_service()

        # Find the repository by URL
        repo_id = await repo_query_service.find_repo_by_url(repo_url)
        if not repo_id:
            raise ValueError(f"Repository not found: {repo_url}")

        if not commit_sha:
            commit_sha = await repo_query_service.find_latest_commit(
                repo_id=repo_id,
                max_commits_to_check=100,
            )
            if not commit_sha:
                raise ValueError(
                    f"No commits with database schema found for repository: {repo_url}"
                )

        enrichment_service = mcp_context.server_factory.enrichment_query_service()
        enrichments = await enrichment_service.get_database_schema_for_commit(
            commit_sha
        )
        return _format_enrichments(enrichments)

    @mcp_server.tool()
    async def get_cookbook(
        ctx: Context,
        repo_url: Annotated[
            str,
            Field(description="The repository URL (e.g., github.com/user/repo)"),
        ],
        commit_sha: Annotated[
            str | None,
            Field(
                description=(
                    "Optional commit SHA. If not provided, uses most recent "
                    "commit with cookbook examples"
                )
            ),
        ] = None,
    ) -> str:
        """Get curated cookbook-style code examples showing common usage patterns.

        Use this to understand:
        - How-to guides for common tasks
        - Working examples with full context
        - Best practices and recommended patterns
        - Step-by-step usage instructions

        Prefer this over search() when you need complete, contextual examples of how
        to accomplish specific tasks rather than searching for code fragments.

        **Important**: If you don't know which repositories are available, call
        list_repositories() first to see the available options.
        """
        mcp_context: MCPContext = ctx.request_context.lifespan_context
        repo_query_service = mcp_context.server_factory.repository_query_service()

        # Find the repository by URL
        repo_id = await repo_query_service.find_repo_by_url(repo_url)
        if not repo_id:
            raise ValueError(f"Repository not found: {repo_url}")

        if not commit_sha:
            commit_sha = await repo_query_service.find_latest_commit(
                repo_id=repo_id,
                max_commits_to_check=100,
            )
            if not commit_sha:
                raise ValueError(
                    f"No commits with cookbook found for repository: {repo_url}"
                )

        enrichment_service = mcp_context.server_factory.enrichment_query_service()
        enrichments = await enrichment_service.get_cookbook_for_commit(commit_sha)
        return _format_enrichments(enrichments)


# FastAPI-integrated MCP server
mcp = create_mcp_server(
    name="Kodit",
    instructions=(
        "This server provides access to code knowledge through multiple "
        "complementary tools:\n\n"
        "**Discovery workflow:**\n"
        "1. Use list_repositories() first to see available repositories\n"
        "2. Then use repository-specific tools with the discovered repo URLs\n\n"
        "**Available tools:**\n"
        "- list_repositories() - Discover available repositories (call this first!)\n"
        "- get_architecture_docs() - High-level structure and design\n"
        "- get_api_docs() - Interface documentation\n"
        "- get_commit_description() - Recent changes and context\n"
        "- get_database_schema() - Data models\n"
        "- get_cookbook() - Complete usage examples\n"
        "- search() - Find specific code snippets matching keywords\n\n"
        "Choose the most appropriate tool based on what information you need. "
        "Often starting with architecture or API docs provides better context than "
        "immediately searching for code snippets."
    ),
)

# Register the MCP tools
register_mcp_tools(mcp)


def create_stdio_mcp_server() -> None:
    """Create and run a STDIO MCP server for kodit."""
    mcp.run(transport="stdio", show_banner=False)



"""Documentation parser protocol and implementations."""

import re
from abc import ABC, abstractmethod

from kodit.infrastructure.example_extraction.code_block import CodeBlock


class DocumentationParser(ABC):
    """Protocol for parsing code blocks from documentation."""

    @abstractmethod
    def parse(self, content: str) -> list[CodeBlock]:
        """Extract code blocks from documentation content."""


class MarkdownParser(DocumentationParser):
    """Parser for Markdown documentation."""

    def parse(self, content: str) -> list[CodeBlock]:
        """Extract code blocks from Markdown content."""
        blocks: list[CodeBlock] = []
        lines = content.split("\n")
        i = 0

        while i < len(lines):
            line = lines[i]
            match = re.match(r"^```(\w+)?", line)
            if match:
                language = match.group(1)
                line_start = i + 1
                code_lines = []
                i += 1

                while i < len(lines) and not lines[i].startswith("```"):
                    code_lines.append(lines[i])
                    i += 1

                if code_lines:
                    context = self._find_context(lines, line_start - 1)
                    blocks.append(
                        CodeBlock(
                            content="\n".join(code_lines),
                            language=language,
                            line_start=line_start,
                            line_end=i - 1,
                            context=context,
                        )
                    )
            i += 1

        return blocks

    def _find_context(self, lines: list[str], block_line: int) -> str | None:
        """Find heading or preceding paragraph near code block."""
        heading = None
        for i in range(max(0, block_line - 10), block_line):
            if lines[i].startswith("#"):
                heading = lines[i].lstrip("#").strip()

        if heading:
            return heading

        paragraph_lines = []
        for i in range(max(0, block_line - 3), block_line):
            stripped = lines[i].strip()
            if stripped and not stripped.startswith("#"):
                paragraph_lines.append(stripped)

        return " ".join(paragraph_lines) if paragraph_lines else None


class RstParser(DocumentationParser):
    """Parser for reStructuredText documentation."""

    def parse(self, content: str) -> list[CodeBlock]:
        """Extract code blocks from RST content."""
        blocks: list[CodeBlock] = []
        lines = content.split("\n")
        i = 0

        while i < len(lines):
            line = lines[i]
            match = re.match(r"^\.\.\s+(code-block|code)::\s*(\w+)?", line)
            if match:
                language = match.group(2)
                i += 1

                while i < len(lines) and not lines[i].strip():
                    i += 1

                if i >= len(lines):
                    break

                base_indent = len(lines[i]) - len(lines[i].lstrip())
                line_start = i
                code_lines = []

                while i < len(lines):
                    current_line = lines[i]
                    if not current_line.strip():
                        i += 1
                        continue
                    current_indent = len(current_line) - len(current_line.lstrip())
                    if current_indent < base_indent:
                        break
                    code_lines.append(current_line[base_indent:])
                    i += 1

                if code_lines:
                    blocks.append(
                        CodeBlock(
                            content="\n".join(code_lines),
                            language=language,
                            line_start=line_start,
                            line_end=i - 1,
                            context=None,
                        )
                    )
            else:
                i += 1

        return blocks


class ParserFactory:
    """Factory for creating documentation parsers."""

    @staticmethod
    def create(file_extension: str) -> DocumentationParser | None:
        """Create parser based on file extension."""
        extension = file_extension.lower()
        if extension in {".md", ".markdown"}:
            return MarkdownParser()
        if extension in {".rst"}:
            return RstParser()
        return None



"""End-to-end test for Python API documentation generation."""

import tempfile
from datetime import UTC, datetime
from pathlib import Path

from kodit.domain.entities.git import GitFile
from kodit.infrastructure.slicing.api_doc_extractor import APIDocExtractor


def test_python_class_with_constructor_and_method() -> None:
    """Test Python class with constructor parameter and method parameter.

    This test verifies the end-to-end flow of:
    1. Parsing a Python class with a constructor that has parameters
    2. Parsing a method with parameters
    3. Generating Pydoc-Markdown style documentation
    4. Verifying the output format includes inline constructor parameters
    """
    # Create a test Python file with a simple class
    python_code = '''"""Calculator module for basic math operations."""


class Calculator:
    """A simple calculator class.

    This calculator performs basic arithmetic operations.
    """

    def __init__(
        self,
        precision: int = 2,
    ):
        """Initialize the calculator.

        Args:
            precision: Number of decimal places for results
        """
        self.precision = precision

    def add(
        self,
        a: float,
        b: float,
    ) -> float:
        """Add two numbers together.

        Args:
            a: First number
            b: Second number

        Returns:
            The sum of a and b
        """
        return round(a + b, self.precision)
'''

    # Write to a temporary file
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as temp_file:
        temp_file.write(python_code)
        temp_path = temp_file.name

    try:
        # Create GitFile for the test file
        git_file = GitFile(
            created_at=datetime.now(tz=UTC),
            blob_sha="test_calculator_sha",
            commit_sha="test_commit_sha",
            path=temp_path,
            mime_type="text/x-python",
            size=Path(temp_path).stat().st_size,
            extension=".py",
        )

        # Extract API docs
        extractor = APIDocExtractor()
        enrichments = extractor.extract_api_docs([git_file], "python")

        # Verify we got exactly one enrichment
        assert len(enrichments) == 1, f"Expected 1 enrichment, got {len(enrichments)}"

        enrichment = enrichments[0]
        content = enrichment.content

        # Print the generated documentation for inspection
        print("\n" + "=" * 80)  # noqa: T201
        print("GENERATED PYTHON API DOCUMENTATION:")  # noqa: T201
        print("=" * 80)  # noqa: T201
        print(content)  # noqa: T201
        print("=" * 80 + "\n")  # noqa: T201

        # Verify enrichment metadata
        assert enrichment.type == "usage"
        assert enrichment.subtype == "api_docs"
        assert enrichment.language == "python"

        # Verify the documentation structure
        assert "# python API Reference" in content
        assert "## " in content  # Should have module section

        # Verify the class is documented
        assert "Calculator" in content

        # Verify constructor parameters appear inline (Pydoc-Markdown style)
        # The class signature should include constructor params
        assert "precision: int = 2" in content

        # Verify the class docstring is present
        assert "A simple calculator class" in content

        # Verify the method is documented
        assert "add" in content
        assert "#### Methods" in content or "###" in content

        # Verify method parameters are in signature
        assert "a: float" in content
        assert "b: float" in content

        # Verify method docstring
        assert "Add two numbers together" in content

        # Verify code fences use Python syntax highlighting
        assert "```py" in content or "```python" in content

        # Verify __init__ is NOT shown as a separate method
        assert "__init__" not in content or content.count("__init__") == 0

    finally:
        # Clean up temporary file
        Path(temp_path).unlink()


def test_python_class_multiple_methods() -> None:
    """Test Python class with multiple methods to verify complete documentation."""
    python_code = '''"""Math utilities module."""


class MathUtils:
    """Mathematical utility functions."""

    def __init__(self, base: int, multiplier: float = 1.0):
        """Initialize math utilities.

        Args:
            base: Base value for calculations
            multiplier: Multiplier to apply to results
        """
        self.base = base
        self.multiplier = multiplier

    def square(self, x: int) -> int:
        """Calculate the square of a number.

        Args:
            x: Number to square

        Returns:
            The square of x
        """
        return int((x ** 2) * self.multiplier)

    def cube(self, x: int) -> int:
        """Calculate the cube of a number.

        Args:
            x: Number to cube

        Returns:
            The cube of x
        """
        return int((x ** 3) * self.multiplier)
'''

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as temp_file:
        temp_file.write(python_code)
        temp_path = temp_file.name

    try:
        git_file = GitFile(
            created_at=datetime.now(tz=UTC),
            blob_sha="test_mathutils_sha",
            commit_sha="test_commit_sha",
            path=temp_path,
            mime_type="text/x-python",
            size=Path(temp_path).stat().st_size,
            extension=".py",
        )

        extractor = APIDocExtractor()
        enrichments = extractor.extract_api_docs([git_file], "python")

        assert len(enrichments) == 1
        content = enrichments[0].content

        # Print for inspection
        print("\n" + "=" * 80)  # noqa: T201
        print("GENERATED PYTHON API DOCUMENTATION (Multiple Methods):")  # noqa: T201
        print("=" * 80)  # noqa: T201
        print(content)  # noqa: T201
        print("=" * 80 + "\n")  # noqa: T201

        # Verify class with inline constructor params
        assert "MathUtils" in content
        assert "base: int" in content
        assert "multiplier: float = 1.0" in content

        # Verify both methods are documented
        assert "square" in content
        assert "cube" in content

        # Verify method signatures include parameters
        assert "x: int" in content

        # Verify docstrings
        assert "Calculate the square of a number" in content
        assert "Calculate the cube of a number" in content

    finally:
        Path(temp_path).unlink()



# ruff: noqa
"""add task status

Revision ID: b9cd1c3fd762
Revises: 9cf0e87de578
Create Date: 2025-09-05 13:41:29.645898

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "b9cd1c3fd762"
down_revision: Union[str, None] = "9cf0e87de578"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "task_status",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("operation", sa.String(length=255), nullable=False),
        sa.Column("trackable_id", sa.Integer(), nullable=True),
        sa.Column("trackable_type", sa.String(length=255), nullable=True),
        sa.Column("parent", sa.String(length=255), nullable=True),
        sa.Column("message", sa.UnicodeText(), nullable=False),
        sa.Column("state", sa.String(length=255), nullable=False),
        sa.Column("error", sa.UnicodeText(), nullable=False),
        sa.Column("total", sa.Integer(), nullable=False),
        sa.Column("current", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["parent"],
            ["task_status.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_task_status_id"), "task_status", ["id"], unique=False)
    op.create_index(
        op.f("ix_task_status_operation"), "task_status", ["operation"], unique=False
    )
    op.create_index(
        op.f("ix_task_status_parent"), "task_status", ["parent"], unique=False
    )
    op.create_index(
        op.f("ix_task_status_trackable_id"),
        "task_status",
        ["trackable_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_task_status_trackable_type"),
        "task_status",
        ["trackable_type"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_task_status_trackable_type"), table_name="task_status")
    op.drop_index(op.f("ix_task_status_trackable_id"), table_name="task_status")
    op.drop_index(op.f("ix_task_status_parent"), table_name="task_status")
    op.drop_index(op.f("ix_task_status_operation"), table_name="task_status")
    op.drop_index(op.f("ix_task_status_id"), table_name="task_status")
    op.drop_table("task_status")
    # ### end Alembic commands ###



"""Dump the OpenAPI json schema to a file."""

import argparse
import json
from pathlib import Path
from typing import Any

from openapi_markdown.generator import to_markdown  # type: ignore[import-untyped]
from uvicorn.importer import import_from_string

parser = argparse.ArgumentParser(prog="dump-openapi.py")
parser.add_argument(
    "app", help='App import string. Eg. "kodit.app:app"', default="kodit.app:app"
)
parser.add_argument("--out-dir", help="Output directory", default="docs/reference/api")

if __name__ == "__main__":
    args = parser.parse_args()

    app = import_from_string(args.app)
    openapi = app.openapi()

    git_tag = openapi["info"]["version"]
    if not git_tag:
        raise ValueError(f"Invalid version: {openapi['info']}")
    # Strip any rcxxx suffix
    git_tag = git_tag.split("rc")[0]
    openapi["info"]["version"] = git_tag

    output_json_file = Path(args.out_dir) / "openapi.json"

    with output_json_file.open("w") as f:
        json.dump(openapi, f, indent=2)

    output_md_file = Path(args.out_dir) / "index.md"
    templates_dir = Path(args.out_dir) / "templates"
    options: dict[str, Any] = {}

    to_markdown(str(output_json_file), str(output_md_file), str(templates_dir), options)



"""Handler for creating summary embeddings for examples in a commit."""

from typing import TYPE_CHECKING, Any

from kodit.application.services.reporting import ProgressTracker
from kodit.domain.enrichments.enrichment import EnrichmentV2
from kodit.domain.protocols import (
    EnrichmentAssociationRepository,
    EnrichmentV2Repository,
)
from kodit.domain.services.embedding_service import EmbeddingDomainService
from kodit.domain.value_objects import (
    Document,
    IndexRequest,
    TaskOperation,
    TrackableType,
)
from kodit.infrastructure.sqlalchemy import entities as db_entities
from kodit.infrastructure.sqlalchemy.embedding_repository import (
    SqlAlchemyEmbeddingRepository,
)
from kodit.infrastructure.sqlalchemy.entities import EmbeddingType
from kodit.infrastructure.sqlalchemy.query import (
    EnrichmentAssociationQueryBuilder,
    FilterOperator,
    QueryBuilder,
)

if TYPE_CHECKING:
    from kodit.application.services.enrichment_query_service import (
        EnrichmentQueryService,
    )


class CreateExampleSummaryEmbeddingsHandler:
    """Handler for creating summary embeddings for examples."""

    def __init__(  # noqa: PLR0913
        self,
        text_search_service: EmbeddingDomainService,
        embedding_repository: SqlAlchemyEmbeddingRepository,
        enrichment_v2_repository: EnrichmentV2Repository,
        enrichment_association_repository: EnrichmentAssociationRepository,
        enrichment_query_service: "EnrichmentQueryService",
        operation: ProgressTracker,
    ) -> None:
        """Initialize the create example summary embeddings handler."""
        self.text_search_service = text_search_service
        self.embedding_repository = embedding_repository
        self.enrichment_v2_repository = enrichment_v2_repository
        self.enrichment_association_repository = enrichment_association_repository
        self.enrichment_query_service = enrichment_query_service
        self.operation = operation

    async def execute(self, payload: dict[str, Any]) -> None:
        """Execute create example summary embeddings operation."""
        repository_id = payload["repository_id"]
        commit_sha = payload["commit_sha"]

        async with self.operation.create_child(
            TaskOperation.CREATE_EXAMPLE_SUMMARY_EMBEDDINGS_FOR_COMMIT,
            trackable_type=TrackableType.KODIT_REPOSITORY,
            trackable_id=repository_id,
        ) as step:
            all_example_enrichments = (
                await self.enrichment_query_service.get_all_examples_for_commit(
                    commit_sha
                )
            )
            if not all_example_enrichments:
                await step.skip("No examples to create summary embeddings")
                return

            query = EnrichmentAssociationQueryBuilder.for_enrichment_associations(
                entity_type=db_entities.EnrichmentV2.__tablename__,
                entity_ids=[
                    str(example.id) for example in all_example_enrichments if example.id
                ],
            )
            summary_enrichment_associations = (
                await self.enrichment_association_repository.find(query)
            )

            if not summary_enrichment_associations:
                await step.skip("No summary enrichments found for examples")
                return

            summary_enrichments = await self.enrichment_v2_repository.find(
                QueryBuilder().filter(
                    "id",
                    FilterOperator.IN,
                    [
                        association.enrichment_id
                        for association in summary_enrichment_associations
                    ],
                )
            )

            new_summaries = await self._new_examples_for_type(
                summary_enrichments, EmbeddingType.TEXT
            )
            if not new_summaries:
                await step.skip("All examples already have text embeddings")
                return

            await step.set_total(len(new_summaries))
            processed = 0

            documents_with_summaries = [
                Document(snippet_id=str(summary.id), text=summary.content)
                for summary in new_summaries
                if summary.id
            ]

            async for result in self.text_search_service.index_documents(
                IndexRequest(documents=documents_with_summaries)
            ):
                processed += len(result)
                await step.set_current(
                    processed, "Creating text embeddings for example summaries"
                )

    async def _new_examples_for_type(
        self, all_examples: list[EnrichmentV2], embedding_type: EmbeddingType
    ) -> list[EnrichmentV2]:
        """Get new examples for a given type."""
        existing_embeddings = (
            await self.embedding_repository.list_embeddings_by_snippet_ids_and_type(
                [str(e.id) for e in all_examples], embedding_type
            )
        )
        if existing_embeddings:
            return []
        existing_embeddings_by_example_id = {
            embedding.snippet_id: embedding for embedding in existing_embeddings
        }
        return [
            e for e in all_examples if e.id not in existing_embeddings_by_example_id
        ]



"""Cookbook enrichment entity."""

from dataclasses import dataclass

from kodit.domain.enrichments.usage.usage import UsageEnrichment

ENRICHMENT_SUBTYPE_COOKBOOK = "cookbook"


@dataclass(frozen=True)
class CookbookEnrichment(UsageEnrichment):
    """Cookbook enrichment containing usage examples for a repository."""

    @property
    def subtype(self) -> str | None:
        """Return the enrichment subtype."""
        return ENRICHMENT_SUBTYPE_COOKBOOK



"""Code block extraction from documentation."""

from dataclasses import dataclass


@dataclass(frozen=True)
class CodeBlock:
    """Represents a code block extracted from documentation."""

    content: str
    language: str | None
    line_start: int
    line_end: int
    context: str | None = None



"""Search router for the REST API."""

from fastapi import APIRouter, HTTPException

from kodit.domain.value_objects import MultiSearchRequest, SnippetSearchFilters
from kodit.infrastructure.api.v1.dependencies import (
    CodeSearchAppServiceDep,
    RepositoryQueryServiceDep,
)
from kodit.infrastructure.api.v1.schemas.search import (
    SearchRequest,
    SearchResponse,
    SnippetAttributes,
    SnippetData,
)
from kodit.infrastructure.api.v1.schemas.snippet import (
    EnrichmentSchema,
    GitFileSchema,
    SnippetContentSchema,
)

router = APIRouter(tags=["search"])


@router.post("/api/v1/search")
async def search_snippets(
    request: SearchRequest,
    search_application_service: CodeSearchAppServiceDep,
    repository_query_service: RepositoryQueryServiceDep,
) -> SearchResponse:
    """Search code snippets with filters matching MCP tool."""
    # Validate source_repo if provided
    if request.sources:
        source_repo = request.sources[0]
        repo_id = await repository_query_service.find_repo_by_url(source_repo)
        if not repo_id:
            raise HTTPException(
                status_code=404, detail=f"Repository not found: {source_repo}"
            )

    # Convert API request to domain request
    domain_request = MultiSearchRequest(
        keywords=request.data.attributes.keywords,
        code_query=request.data.attributes.code,
        text_query=request.data.attributes.text,
        top_k=request.limit or 10,
        filters=SnippetSearchFilters(
            language=request.languages[0] if request.languages else None,
            author=request.authors[0] if request.authors else None,
            created_after=request.start_date,
            created_before=request.end_date,
            source_repo=request.sources[0] if request.sources else None,
            file_path=request.file_patterns[0] if request.file_patterns else None,
            enrichment_types=request.enrichment_types,
            enrichment_subtypes=request.enrichment_subtypes,
            commit_sha=request.commit_sha,
        )
        if any(
            [
                request.languages,
                request.authors,
                request.start_date,
                request.end_date,
                request.sources,
                request.file_patterns,
                request.enrichment_types,
                request.enrichment_subtypes,
                request.commit_sha,
            ]
        )
        else None,
    )

    # Execute search using application service
    results = await search_application_service.search(domain_request)

    return SearchResponse(
        data=[
            SnippetData(
                type=result.enrichment_subtype or result.enrichment_type,
                id=result.snippet.id,
                attributes=SnippetAttributes(
                    created_at=result.snippet.created_at,
                    updated_at=result.snippet.updated_at,
                    derives_from=[
                        GitFileSchema(
                            blob_sha=file.blob_sha,
                            path=file.path,
                            mime_type=file.mime_type,
                            size=file.size,
                        )
                        for file in result.snippet.derives_from
                    ],
                    content=SnippetContentSchema(
                        value=result.snippet.content,
                        language=result.snippet.extension,
                    ),
                    enrichments=[
                        EnrichmentSchema(
                            type=enrichment.type,
                            content=enrichment.content,
                        )
                        for enrichment in result.snippet.enrichments
                    ],
                    original_scores=result.original_scores,
                ),
            )
            for result in results
        ]
    )



"""Handler for generating cookbook examples for a commit."""

from collections import defaultdict
from typing import TYPE_CHECKING, Any

import structlog

from kodit.application.services.reporting import ProgressTracker
from kodit.domain.enrichments.enricher import Enricher
from kodit.domain.enrichments.enrichment import CommitEnrichmentAssociation
from kodit.domain.enrichments.request import (
    EnrichmentRequest as GenericEnrichmentRequest,
)
from kodit.domain.enrichments.usage.cookbook import CookbookEnrichment
from kodit.domain.protocols import (
    EnrichmentAssociationRepository,
    EnrichmentV2Repository,
    GitFileRepository,
    GitRepoRepository,
)
from kodit.domain.services.cookbook_context_service import (
    COOKBOOK_SYSTEM_PROMPT,
    COOKBOOK_TASK_PROMPT,
    CookbookContextService,
)
from kodit.domain.value_objects import LanguageMapping, TaskOperation, TrackableType
from kodit.infrastructure.sqlalchemy.query import GitFileQueryBuilder

if TYPE_CHECKING:
    from kodit.application.services.enrichment_query_service import (
        EnrichmentQueryService,
    )
    from kodit.domain.entities.git import GitFile


class CookbookHandler:
    """Handler for generating cookbook examples for a commit."""

    def __init__(  # noqa: PLR0913
        self,
        repo_repository: GitRepoRepository,
        git_file_repository: GitFileRepository,
        cookbook_context_service: CookbookContextService,
        enricher_service: Enricher,
        enrichment_v2_repository: EnrichmentV2Repository,
        enrichment_association_repository: EnrichmentAssociationRepository,
        enrichment_query_service: "EnrichmentQueryService",
        operation: ProgressTracker,
    ) -> None:
        """Initialize the cookbook handler."""
        self.repo_repository = repo_repository
        self.git_file_repository = git_file_repository
        self.cookbook_context_service = cookbook_context_service
        self.enricher_service = enricher_service
        self.enrichment_v2_repository = enrichment_v2_repository
        self.enrichment_association_repository = enrichment_association_repository
        self.enrichment_query_service = enrichment_query_service
        self.operation = operation
        self._log = structlog.get_logger(__name__)

    async def execute(self, payload: dict[str, Any]) -> None:
        """Execute cookbook generation operation."""
        repository_id = payload["repository_id"]
        commit_sha = payload["commit_sha"]

        async with self.operation.create_child(
            TaskOperation.CREATE_COOKBOOK_FOR_COMMIT,
            trackable_type=TrackableType.KODIT_REPOSITORY,
            trackable_id=repository_id,
        ) as step:
            # Check if cookbook already exists for this commit
            if await self.enrichment_query_service.has_cookbook_for_commit(commit_sha):
                await step.skip("Cookbook already exists for commit")
                return

            # Get repository path
            repo = await self.repo_repository.get(repository_id)
            if not repo.cloned_path:
                raise ValueError(f"Repository {repository_id} has never been cloned")

            await step.set_total(4)
            await step.set_current(1, "Getting files for cookbook generation")

            # Get files for the commit
            files = await self.git_file_repository.find(
                GitFileQueryBuilder().for_commit_sha(commit_sha)
            )
            if not files:
                await step.skip("No files to generate cookbook from")
                return

            # Group files by language and find primary language
            lang_files_map: dict[str, list[GitFile]] = defaultdict(list)
            for file in files:
                try:
                    lang = LanguageMapping.get_language_for_extension(file.extension)
                except ValueError:
                    continue
                lang_files_map[lang].append(file)

            if not lang_files_map:
                await step.skip("No supported languages found for cookbook")
                return

            # Use the language with the most files as primary
            primary_lang = max(lang_files_map.items(), key=lambda x: len(x[1]))[0]
            primary_lang_files = lang_files_map[primary_lang]

            await step.set_current(2, f"Parsing {primary_lang} code with AST")

            # Parse API structure using AST analyzer
            api_modules = None
            try:
                from kodit.infrastructure.slicing.ast_analyzer import ASTAnalyzer

                analyzer = ASTAnalyzer(primary_lang)
                parsed_files = analyzer.parse_files(primary_lang_files)
                api_modules = analyzer.extract_module_definitions(
                    parsed_files, include_private=False
                )
                # Filter out test modules
                api_modules = [
                    m
                    for m in api_modules
                    if not self._is_test_module_path(m.module_path)
                ]
            except (ValueError, Exception) as e:
                self._log.debug(
                    "Could not parse API structure, continuing without it",
                    language=primary_lang,
                    error=str(e),
                )

            await step.set_current(3, "Gathering repository context for cookbook")

            # Gather context for cookbook generation
            repository_context = await self.cookbook_context_service.gather_context(
                repo.cloned_path, language=primary_lang, api_modules=api_modules
            )

            await step.set_current(4, "Generating cookbook examples with LLM")

            # Generate cookbook through the enricher
            enrichment_request = GenericEnrichmentRequest(
                id=commit_sha,
                text=COOKBOOK_TASK_PROMPT.format(repository_context=repository_context),
                system_prompt=COOKBOOK_SYSTEM_PROMPT,
            )

            enriched_content = ""
            async for response in self.enricher_service.enrich([enrichment_request]):
                enriched_content = response.text

            # Create and save cookbook enrichment
            enrichment = await self.enrichment_v2_repository.save(
                CookbookEnrichment(
                    content=enriched_content,
                )
            )
            if not enrichment or not enrichment.id:
                raise ValueError(
                    f"Failed to save cookbook enrichment for commit {commit_sha}"
                )
            await self.enrichment_association_repository.save(
                CommitEnrichmentAssociation(
                    enrichment_id=enrichment.id,
                    entity_id=commit_sha,
                )
            )

    def _is_test_module_path(self, module_path: str) -> bool:
        """Check if a module path appears to be a test module."""
        module_path_lower = module_path.lower()
        test_indicators = ["test", "tests", "__tests__", "_test", "spec"]
        return any(indicator in module_path_lower for indicator in test_indicators)



"""Tests for the LiteLLM embedding provider."""

from typing import Any
from unittest.mock import AsyncMock

import pytest

from kodit.config import Endpoint
from kodit.domain.value_objects import EmbeddingRequest
from kodit.infrastructure.embedding.embedding_providers.litellm_embedding_provider import (  # noqa: E501
    LiteLLMEmbeddingProvider,
)


class TestLiteLLMEmbeddingProvider:
    """Test the LiteLLM embedding provider."""

    def test_init_default_values(self) -> None:
        """Test initialization with default values."""
        endpoint = Endpoint()
        provider = LiteLLMEmbeddingProvider(endpoint)
        assert provider.endpoint.model is None
        assert provider.endpoint.api_key is None
        assert provider.endpoint.base_url is None
        assert provider.endpoint.timeout == 60
        assert provider.endpoint.extra_params is None
        assert provider.log is not None

    def test_init_custom_values(self) -> None:
        """Test initialization with custom values."""
        extra_params = {"temperature": 0.5}
        endpoint = Endpoint(
            model="text-embedding-3-large",
            api_key="test-api-key",
            base_url="https://custom.openai.com",
            timeout=60.0,
            extra_params=extra_params,
        )
        provider = LiteLLMEmbeddingProvider(endpoint)
        assert provider.endpoint.model == "text-embedding-3-large"
        assert provider.endpoint.api_key == "test-api-key"
        assert provider.endpoint.base_url == "https://custom.openai.com"
        assert provider.endpoint.timeout == 60.0
        assert provider.endpoint.extra_params == extra_params

    @pytest.mark.asyncio
    async def test_embed_empty_requests(self) -> None:
        """Test embedding with empty requests."""
        endpoint = Endpoint()
        provider = LiteLLMEmbeddingProvider(endpoint)

        results = []
        async for batch in provider.embed([]):
            results.extend(batch)

        assert len(results) == 0

    @pytest.mark.asyncio
    async def test_embed_single_request_success(self) -> None:
        """Test successful embedding with a single request."""
        endpoint = Endpoint(model="text-embedding-3-small")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1, 0.2, 0.3, 0.4, 0.5] * 300}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="python programming")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        assert len(results) == 1
        assert results[0].snippet_id == "1"
        assert len(results[0].embedding) == 1500
        assert all(isinstance(v, float) for v in results[0].embedding)

        # Verify the provider's embedding method was called correctly
        provider.provider.embedding.assert_called_once_with(["python programming"])

    @pytest.mark.asyncio
    async def test_embed_multiple_requests_success(self) -> None:
        """Test successful embedding with multiple requests."""
        endpoint = Endpoint(model="text-embedding-3-small")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={
                "data": [
                    {"embedding": [0.1, 0.2, 0.3] * 500},  # 1500 dims
                    {"embedding": [0.4, 0.5, 0.6] * 500},  # 1500 dims
                    {"embedding": [0.7, 0.8, 0.9] * 500},  # 1500 dims
                ]
            }
        )

        requests = [
            EmbeddingRequest(snippet_id="1", text="python programming"),
            EmbeddingRequest(snippet_id="2", text="javascript development"),
            EmbeddingRequest(snippet_id="3", text="java enterprise"),
        ]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        assert len(results) == 3
        for i, result in enumerate(results):
            assert result.snippet_id == str(i + 1)
            assert len(result.embedding) == 1500
            assert all(isinstance(v, float) for v in result.embedding)

        # Verify the provider's embedding method was called correctly
        provider.provider.embedding.assert_called_once_with(
            [
                "python programming",
                "javascript development",
                "java enterprise",
            ]
        )

    @pytest.mark.asyncio
    async def test_embed_with_base_url(self) -> None:
        """Test embedding with custom base URL."""
        endpoint = Endpoint(
            model="text-embedding-3-small", base_url="https://custom.api.com"
        )
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1] * 1500}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="test")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        # Verify the provider's embedding method was called
        provider.provider.embedding.assert_called_once_with(["test"])

    @pytest.mark.asyncio
    async def test_embed_with_api_key(self) -> None:
        """Test embedding with API key."""
        endpoint = Endpoint(model="text-embedding-3-small", api_key="sk-test-key-123")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1] * 1500}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="test")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        # Verify the provider's embedding method was called
        provider.provider.embedding.assert_called_once_with(["test"])

    @pytest.mark.asyncio
    async def test_embed_with_extra_params(self) -> None:
        """Test embedding with extra parameters."""
        extra_params = {"temperature": 0.5, "max_tokens": 100}
        endpoint = Endpoint(model="text-embedding-3-small", extra_params=extra_params)
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1] * 1500}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="test")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        # Verify the provider's embedding method was called
        provider.provider.embedding.assert_called_once_with(["test"])

    @pytest.mark.asyncio
    async def test_embed_batch_processing(self) -> None:
        """Test that requests are processed in batches."""
        endpoint = Endpoint()
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock responses for different batches
        async def mock_embedding_func(texts: list[str]) -> dict[str, Any]:
            return {
                "data": [{"embedding": [0.1] * 1500} for _ in range(len(texts))]
            }

        provider.provider.embedding = AsyncMock(side_effect=mock_embedding_func)  # type: ignore[method-assign]

        # Create more than batch_size requests (batch_size = 10)
        requests = [
            EmbeddingRequest(snippet_id=str(i), text=f"text {i}") for i in range(15)
        ]

        batch_count = 0
        total_results = []
        async for batch in provider.embed(requests):
            batch_count += 1
            total_results.extend(batch)

        assert len(total_results) == 15
        assert batch_count == 2  # Should be 2 batches: 10 + 5
        assert provider.provider.embedding.call_count == 2

    @pytest.mark.asyncio
    async def test_embed_api_error_handling(self) -> None:
        """Test handling of API errors."""
        endpoint = Endpoint(model="text-embedding-3-small")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method to raise an error
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            side_effect=Exception("LiteLLM API Error")
        )

        requests = [EmbeddingRequest(snippet_id="1", text="python programming")]

        # Should raise exception on error
        with pytest.raises(Exception, match="LiteLLM API Error"):
            async for _ in provider.embed(requests):
                pass

    @pytest.mark.asyncio
    async def test_embed_response_without_model_dump(self) -> None:
        """Test handling response without model_dump method."""
        endpoint = Endpoint()
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock response that doesn't have model_dump method (dict response)
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1] * 1500}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="test")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        assert len(results) == 1
        assert results[0].snippet_id == "1"
        assert len(results[0].embedding) == 1500

    @pytest.mark.asyncio
    async def test_embed_custom_model(self) -> None:
        """Test embedding with a custom model."""
        endpoint = Endpoint(model="claude-3-haiku-20240307")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Mock the provider's embedding method
        provider.provider.embedding = AsyncMock(  # type: ignore[method-assign]
            return_value={"data": [{"embedding": [0.1] * 1500}]}
        )

        requests = [EmbeddingRequest(snippet_id="1", text="test text")]

        results = []
        async for batch in provider.embed(requests):
            results.extend(batch)

        # Verify the provider's embedding method was called
        provider.provider.embedding.assert_called_once_with(["test text"])

    @pytest.mark.asyncio
    async def test_socket_path_setup(self) -> None:
        """Test Unix socket setup."""
        endpoint = Endpoint(socket_path="/var/run/test.sock")
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Verify socket_path was stored
        assert provider.endpoint.socket_path == "/var/run/test.sock"

        # Should complete without error
        await provider.close()

    @pytest.mark.asyncio
    async def test_socket_path_httpx_client_setup(self) -> None:
        """Test that Unix socket creates proper HTTPX client."""
        endpoint = Endpoint(socket_path="/var/run/test.sock", timeout=60.0)
        provider = LiteLLMEmbeddingProvider(endpoint)

        # Verify socket_path was passed to the provider
        assert provider.endpoint.socket_path == "/var/run/test.sock"
        assert provider.endpoint.timeout == 60.0

        await provider.close()

    @pytest.mark.asyncio
    async def test_close(self) -> None:
        """Test close method (should not raise any errors)."""
        endpoint = Endpoint()
        provider = LiteLLMEmbeddingProvider(endpoint)
        # Should complete without error
        await provider.close()



"""Schemas for the application context."""

from typing import TypedDict

from kodit.application.factories.server_factory import ServerFactory
from kodit.config import AppContext


class AppLifespanState(TypedDict):
    """Application lifespan state."""

    app_context: AppContext
    server_factory: ServerFactory
